from __future__ import annotations

import numpy as np
import scipy.sparse
import scipy.sparse.linalg

class NonSquareError(Exception):
    pass

class NonSymmetricError(Exception):
    pass

class Matrix:

    """
    Internal data structure used to represent matrices. Can represent sparse, dense and diagonal matrices.

    Input
    -----
    A                   numpy 2D array
                        or scipy sparse array
                        or (N,) vector          rerpesents dia(vector)
                        or scalar               represents scalar * identity matrix of any size.
    assert_square       bool
    assert_symmetric    bool

    Methods:
    -------
    get_shape()                             returns matrix shape.
    is_diagonal()                           True if A is a diagonal matrix
    is_square()                             True if A is a square matrix
    diag()                                  returns diagonal of A. (if A is scalar; returns scalar)
    select(mask)                            returns A[mask, :][:, mask]
    stack(matrix: Matrix)                   returns [[self, 0], [0, matrix]]
    """

    def __init__(self,  A, assert_square=False, assert_symmetric=False):
        self.A = np.array(A) if not scipy.sparse.issparse(A) else A
        if self.A.ndim > 2:
            raise ValueError("matrix must be 2D")
        self.shape = None if self.A.ndim == 0 else self.A.shape
        if scipy.sparse.issparse(self.A):
            if self.shape[0] == self.shape[1]:
                if (self.A - scipy.sparse.diags(self.A.diagonal(), 0)).nnz == 0:
                    self.A = self.A.diagonal()
        if self.A.ndim == 1:
            if np.allclose(self.A, self.A[0]):
                self.A = self.A[0]
        self.is_scalar = self.A.ndim == 0
        if assert_square:
            if not self.is_diagonal():
                if not self.is_square():
                    raise NonSquareError("Matrix not square")
        if assert_symmetric:
            if not self.is_diagonal():
                if not self._is_symmetric(self.A):
                    raise NonSymmetricError("Matrix not symmetric")
        self.factorization = None

    def get_shape(self):
        if self.shape is None:
            return None
        if self.is_diagonal():
            return (self.shape[0], self.shape[0])
        return self.shape

    def matrix(self, N=None):
        if not self.is_diagonal():
            return self.A
        else:
            return scipy.sparse.diags(self.diag(force_as_vector=True, vector_length=N), 0)

    def is_diagonal(self):
        return self.A.ndim <= 1

    def is_zero(self):
        if self.is_diagonal():
            return np.all(self.diag() == 0)
        return False

    def is_square(self):
        shape = self.get_shape()
        return shape[0] == shape[1] if shape is not None else True

    def diag(self, force_as_vector=False, vector_length=None):
        out = self.A
        if not self.is_diagonal():
            out = self.A.diagonal()
        if not self.is_square():
            raise NonSquareError("Cannot extract diagonal; Matrix not square")
        return np.broadcast_to(out, (self.get_shape()[0] if vector_length is None else vector_length)) if force_as_vector else out

    def select(self, mask) -> Matrix:
        # returns A[mask, :][:, mask]
        if self.is_scalar:
            return Matrix(self.A)
        if self.is_diagonal():
            return Matrix(self.A[mask])
        return Matrix(self.A[mask, :][:, mask])

    def stack(self, matrix: Matrix) -> Matrix:
        # returns [[self, 0], [0, matrix]]
        if self.shape is None or matrix.shape is None:
            raise ValueError("cannot stack matrices of indeterminate size")
        N1, N2 = self.shape[0], matrix.shape[0]
        x1, x2 = self.A, matrix.A
        if self.is_scalar and matrix.is_scalar:
            if self.A == matrix.A:
                return Matrix(x1)
        if self.is_scalar:
            x1 = np.ones(N1, x1.dtype) * x1
        if matrix.is_scalar:
            x2 = np.ones(N2, x2.dtype) * x2
        if self.is_diagonal() and matrix.is_diagonal():
            return Matrix(np.append(x1, x2))
        if self.is_diagonal():
            x1 = scipy.sparse.diags(x1, 0)
        if matrix.is_diagonal():
            x2 = scipy.sparse.diags(x2, 0)
        return Matrix(scipy.sparse.block_diag((x1, x2)))

    def _is_symmetric(self, A):
        if scipy.sparse.isspmatrix(A):
            return (A - A.T).nnz == 0
        else:
            return np.all(A == A.T)

    def __str__(self):
        if self.is_scalar:
            return f"Matrix object representing {self.A} times identity matrix"
        if self.is_diagonal():
            return f"diagonal Matrix object of shape {self.get_shape()}"
        return f"Matrix object of shape {self.get_shape()}"



import time
import numpy as np
import scipy
import scipy.sparse.linalg
import matplotlib
import matplotlib.pyplot as plt
matplotlib.use("TkAgg")

from josephson_circuit import HoneycombArray, SquareArray
from josephson_circuit import Circuit
from static_problem import lobpcg_test_negative_definite, DefaultCurrentPhaseRelation, StaticProblem

array = SquareArray(10, 10)
array.set_inductance_factors(1)
Is = array.horizontal_junctions()
f = 0.01

n = np.zeros(array.face_count())
n[array.locate_faces(4.5, 4.5)] = 1

prob = StaticProblem(array, vortex_configuration=n, frustration=f, current_sources=Is)
_, _, conf, _ = prob.compute_maximal_current()


def is_stable_no_inductance(array, theta):
    """
    Determines if a configuration on an array with no inductance is stable in the sense that
    the Jacobian is negative definite, which it returns along with a StabilityInfo handle.
    Does not explicitly check if configuration is a stationairy point.
    """
    tic = time.perf_counter()
    cp = DefaultCurrentPhaseRelation()
    Nj, Nnr = array._Nj(), array._Nnr()
    Ic = array._Ic().diag(force_as_vector=True, vector_length=Nj)
    d_theta = cp.d_eval(Ic, theta)
    Mr = array._Mr().A
    MrT = Mr.T
    M = array.get_cut_matrix()
    MT = M.T
    A = array.get_cycle_matrix()
    L = array._L().matrix(Nj)
    q = cp.d_eval(Ic, theta)

    J2 = -scipy.sparse.vstack(
        [Mr @ scipy.sparse.diags(q, 0), A @ (scipy.sparse.eye(Nj) + L @ scipy.sparse.diags(q, 0))]).tocsc()

    m = scipy.sparse.vstack(
        [Mr, A @ L]).tocsc()

    J1 = -Mr @ scipy.sparse.diags(d_theta, 0) @ MrT
    preconditioner = scipy.sparse.linalg.LinearOperator((Nnr, Nnr), scipy.sparse.linalg.factorized(J1))

    AL = (A @ L @ A.T).tocoo()
    ALL = scipy.sparse.coo_matrix((AL.data, (AL.row + Nnr, AL.col + Nnr)), shape=(Nj, Nj)).tocsc()
    J3 = - (m @ scipy.sparse.diags(q, 0) @ m.T + ALL)
    select = np.diff(J3.indptr)!=0

    J3 = J3[select, :][:, select]

    print(time.perf_counter() - tic)
    x0 = np.random.rand(Nnr, 1)
    tic = time.perf_counter()
    lobpcg_out = scipy.sparse.linalg.lobpcg(J1, x0, M=preconditioner,
                                            maxiter=1000, tol=1E-10,
                                            retLambdaHistory=True,
                                            retResidualNormsHistory=True)
    print(lobpcg_out[0], time.perf_counter() - tic)

    tic = time.perf_counter()
    w, v = scipy.sparse.linalg.eigs(J1, 1, maxiter=1000, which="LR")
    print(w, time.perf_counter() - tic)


    # tic = time.perf_counter()
    # print(J2.shape)
    # w, v = scipy.sparse.linalg.eigs(J2, 1, M=m, maxiter=1000, which="LR")
    # print(w, time.perf_counter() - tic)


    tic = time.perf_counter()
    print(J2.shape)
    w, v = scipy.sparse.linalg.eigsh(J3, 1, maxiter=1000, which="LA")
    print(w, time.perf_counter() - tic)

    return lobpcg_out


is_stable_no_inductance(array, conf.get_theta())



"""pyJJASim: Josephson Circuit Simulation in Python """


from pyJJAsim.variable_row_array import *
from pyJJAsim.compute import *
from pyJJAsim.embedded_graph import *
from pyJJAsim.josephson_circuit import *
from pyJJAsim.static_problem import *
from pyJJAsim.time_evolution import *
from pyJJAsim.circuit_visualize import *







import numpy as np
import scipy.sparse
import matplotlib
import matplotlib.pyplot as plt
matplotlib.use("TkAgg")

from pyJJAsim.variable_row_array import VarRowArray

class NotSingleComponentError(Exception):
    pass

class NotPlanarEmbeddingError(Exception):
    pass

class NonSimpleError(Exception):
    pass

class SelfLoopError(Exception):
    pass

class NodeNotExistError(Exception):
    pass

class EdgeNotExistError(Exception):
    pass


class EmbeddedGraph:

    """
    Class for embedded 2D graphs. Can be used to construct faces and check planarity.

    Requirements:
     - No self-loops
     - Simple (no multigraph)
    Input:
    x, y            (N,)
    node1, node2    (E,) in range(N)

    Methods:
    get_faces()
    is_planar()
    get_areas()
    get_centroids()
    get_num_components()
    get_boundary_faces()

    """

    def __init__(self, x, y, node1, node2, require_single_component=False,
                 require_planar_embedding=False, _edges_are_sorted=False):
        self.x = np.array(x, dtype=np.double).ravel()
        self.y = np.array(y, dtype=np.double).ravel()
        if len(self.x) != len(self.y):
            raise ValueError("x and y must be same size")
        self.N = len(self.x)
        self.node1, self.node2 = np.array(node1, dtype=int).ravel(), np.array(node2, dtype=int).ravel()
        self.assert_edges_correct_shape()
        self.E = len(self.node1)
        self.assert_edges_contain_existing_nodes()
        self.assert_no_self_loop()

        self.edge_sorter = np.arange(self.edge_count())
        if not _edges_are_sorted:
            self._sort_edges(node1, node2)

        self.assert_nonsimple()
        self.edge_v_array = self._assign_edge_v_array()

        # quantities are computed and stored only when a method needs them.
        self.F = None
        self.faces_v_array = None
        self.face_edges = None
        self.face_nodes = None
        self.face_lengths = None
        self.determinant = None
        self.areas = None
        self.roll_ids = None
        self.boundary_face_indices = None

        if require_single_component:
            self.assert_single_component()
        if require_planar_embedding:
            self.assert_planar_embedding()

    def _assign_edge_v_array(self):
        self.edge_v_array = VarRowArray(np.diff(np.append(np.flatnonzero(np.roll(self.node1, 1) - self.node1), len(self.node1))))
        return self.edge_v_array

    def assert_edges_correct_shape(self):
        if len(self.node1) != len(self.node2):
            raise ValueError("node1 and node2 must be same size")

    def assert_edges_contain_existing_nodes(self):
        if np.any((self.node1 < 0) | (self.node1 >= self.N) | (self.node2 < 0) | (self.node2 >= self.N)):
            raise NodeNotExistError("node1,2 values must be in range(N)")

    def assert_single_component(self):
        if self.get_num_components() != 1:
            raise NotSingleComponentError()

    def assert_planar_embedding(self):
        if not self.is_planar_embedding():
            raise NotPlanarEmbeddingError()

    def assert_no_self_loop(self):
        if np.any(self.node1 == self.node2):
            raise SelfLoopError("no edge is allowed to have identical end nodes.")

    def assert_nonsimple(self):
        if np.any((self.node1[:-1] == self.node1[1:]) & (self.node2[:-1] == self.node2[1:])):
            raise NonSimpleError("no duplicate edges allowed.")

    def get_edges(self):
        return self.node1, self.node2

    def get_edge_ids(self, node1, node2):
        node1, node2 = np.array(node1, dtype=int).ravel(), np.array(node2, dtype=int).ravel()
        mask = node1 < node2
        node1, node2 = np.where(mask, node1, node2), np.where(mask, node2, node1)
        if np.any(node1 > self.node1[-1]):
            raise EdgeNotExistError("queried edge that does not exist")
        starts, ends = self.edge_v_array.row_ranges()
        node1_idx = np.searchsorted(self.node1[starts], node1)
        starts, ends = starts[node1_idx], ends[node1_idx]
        while not np.all(node2 <= self.node2[starts]):
            starts[node2 > self.node2[starts]] += 1
            if not np.all(starts < ends):
                raise EdgeNotExistError("queried edge that does not exist")
        if not np.all((self.node1[starts] == node1) & (self.node2[starts] == node2)):
            raise EdgeNotExistError("queried edge that does not exist")
        return starts

    def add_nodes(self, x, y):
        # if in_place:
        #     self.x = np.append(self.x, np.array(x, dtype=np.double).ravel())
        #     self.y = np.append(self.y, np.array(y, dtype=np.double).ravel())
        #     self.N = len(self.x)
        return EmbeddedGraph(np.append(self.x, np.array(x, dtype=np.double).ravel()), np.append(self.y, np.array(y, dtype=np.double).ravel()), self.node1, self.node2, require_single_component=False)

    def add_edges(self, node1, node2, require_single_component=False, require_planar_embedding=False):
        # if in_place:
        #     self._reset_precomputed_quantities()
        #     self.node1 = np.append(self.node1, np.array(node1, dtype=int).ravel())
        #     self.node2 = np.append(self.node2, np.array(node2, dtype=int).ravel())
        #     self.E = len(self.node1)
        #     self._sort_edges(self.node1, self.node2)
        #     self.assert_edges_correct_shape()
        #     self.assert_edges_contain_existing_nodes()
        #     self.assert_no_self_loop()
        #     self.assert_nonsimple()
        #     self._assign_edge_v_array()
        #     if require_single_component:
        #         self.assert_single_component()
        #     if require_planar_embedding:
        #         self.assert_planar_embedding()
        #     return self
        # else:
        return EmbeddedGraph(self.x, self.y, np.append(self.node1, np.array(node1, dtype=int).ravel()),
                             np.append(self.node2, np.array(node2, dtype=int).ravel()),
                             require_single_component=require_single_component,
                             require_planar_embedding=require_planar_embedding)

    def add_nodes_and_edges(self, x, y, node1, node2, require_single_component=False,
                            require_planar_embedding=False):
        # if in_place:
        #     self._reset_precomputed_quantities()
        #     self.x = np.append(self.x, np.array(x, dtype=np.double).ravel())
        #     self.y = np.append(self.y, np.array(y, dtype=np.double).ravel())
        #     self.N = len(self.x)
        #     self.node1 = np.append(self.node1, np.array(node1, dtype=int).ravel())
        #     self.node2 = np.append(self.node2, np.array(node2, dtype=int).ravel())
        #     self.E = len(self.node1)
        #     self._sort_edges(self.node1, self.node2)
        #     self.assert_edges_correct_shape()
        #     self.assert_edges_contain_existing_nodes()
        #     self.assert_no_self_loop()
        #     self.assert_nonsimple()
        #     self._assign_edge_v_array()
        #     if require_single_component:
        #         self.assert_single_component()
        #     if require_planar_embedding:
        #         self.assert_planar_embedding()
        #     return self
        # else:
        return EmbeddedGraph(np.append(self.x, np.array(x, dtype=np.double).ravel()),
                             np.append(self.y, np.array(y, dtype=np.double).ravel()),
                             np.append(self.node1, np.array(node1, dtype=int).ravel()),
                             np.append(self.node2, np.array(node2, dtype=int).ravel()),
                             require_single_component=require_single_component,
                             require_planar_embedding=require_planar_embedding)

    def remove_nodes(self, node_ids, require_single_component=False, require_planar_embedding=False):
        node_ids = np.array(node_ids, dtype=int).ravel()
        node_map = np.ones(self.node_count(), dtype=int)
        node_map[node_ids] = 0
        node_map = np.cumsum(node_map) - node_map
        node_map[node_ids] = -1
        n1, n2 = node_map[self.node1], node_map[self.node2]
        edge_ids = (n1 >= 0) & (n2 >= 0)
        # if in_place:
        #     self._reset_precomputed_quantities()
        #
        #     edge_map = np.ones(self.edge_count(), dtype=int)
        #     edge_map[edge_ids] = 0
        #     edge_map = np.cumsum(edge_map) - edge_map
        #
        #     self.x = np.delete(self.x, node_ids)
        #     self.y = np.delete(self.y, node_ids)
        #     self.N = len(self.x)
        #     self.node1 = n1[edge_ids]
        #     self.node2 = n2[edge_ids]
        #     self.edge_sorter = edge_map[np.delete(self.edge_sorter, np.argsort(self.edge_sorter)[edge_ids])]
        #     self.E = len(self.node1)
        #
        #     self._assign_edge_v_array()
        #     if require_single_component:
        #         self.assert_single_component()
        #     if require_planar_embedding:
        #         self.assert_planar_embedding()
        #     return self
        # else:
        return EmbeddedGraph(np.delete(self.x, node_ids), np.delete(self.y, node_ids),
                             n1[edge_ids], n2[edge_ids], require_single_component=require_single_component,
                             require_planar_embedding=require_planar_embedding, _edges_are_sorted=True)

    def remove_edges_by_ids(self, edge_ids, require_single_component=False, require_planar_embedding=False):
        # if in_place:
        #     self._reset_precomputed_quantities()
        #
        #     edge_map = np.ones(self.edge_count(), dtype=int)
        #     edge_map[edge_ids] = 0
        #     edge_map = np.cumsum(edge_map) - edge_map
        #
        #     self.node1, self.node2 = np.delete(self.node1, edge_ids), np.delete(self.node2, edge_ids)
        #     self.edge_sorter = edge_map[np.delete(self.edge_sorter, np.argsort(self.edge_sorter)[edge_ids])]
        #
        #     self.E = len(self.node1)
        #     self._sort_edges(self.node1, self.node2)
        #     self._assign_edge_v_array()
        #     if require_single_component:
        #         self.assert_single_component()
        #     if require_planar_embedding:
        #         self.assert_planar_embedding()
        #     return self
        # else:
        return EmbeddedGraph(self.x, self.y, np.delete(self.node1, edge_ids),
                             np.delete(self.node2, edge_ids), _edges_are_sorted=True,
                             require_single_component=require_single_component,
                             require_planar_embedding=require_planar_embedding)

    def remove_edges(self, node1, node2, require_single_component=False, require_planar_embedding=False):
        return self.remove_edges_by_ids(self.get_edge_ids(node1, node2),
                                        require_single_component=require_single_component,
                                        require_planar_embedding=require_planar_embedding)

    def _reset_precomputed_quantities(self):
        self.F = None
        self.faces_v_array = None
        self.face_edges = None
        self.face_nodes = None
        self.face_lengths = None
        self.determinant = None
        self.areas = None
        self.roll_ids = None
        self.boundary_face_indices = None

    def coo(self):
        return self.x, self.y

    def node_count(self):
        return self.N

    def edge_count(self):
        return self.E

    def get_faces(self, include_boundary_faces=True, to_list=False):
        if self.face_edges is None:
            self.face_edges, self.face_lengths = self._construct_faces()
            self.faces_v_array = VarRowArray(self.face_lengths)
            self.F = len(self.face_lengths)
        if include_boundary_faces:
            edges, lengths = self.face_edges, self.face_lengths
        else:
            if self.boundary_face_indices is None:
                self.get_areas()
            edges = self.faces_v_array.delete_rows(self.face_edges, self.boundary_face_indices)
            lengths = np.delete(self.face_lengths, self.boundary_face_indices)
        if to_list:
            return VarRowArray(lengths).to_list(edges)
        return edges, lengths

    def get_face_nodes(self, include_boundary_faces=True, to_list=False):
        if self.face_edges is None:
            self.get_faces()
        if self.face_nodes is None:
            e_idx = self.face_edges % self.edge_count()
            self.face_nodes = np.where(self.face_edges < self.edge_count(), self.node1[e_idx], self.node2[e_idx])
        if include_boundary_faces:
            nodes, lengths = self.face_nodes, self.face_lengths
        else:
            if self.boundary_face_indices is None:
                self.get_areas()
            nodes = self.faces_v_array.delete_rows(self.face_nodes, self.boundary_face_indices)
            lengths = np.delete(self.face_lengths, self.boundary_face_indices)
        if to_list:
            return VarRowArray(lengths).to_list(nodes)
        return nodes, lengths

    def face_count(self, include_boundary_faces=True):
        if self.F is None:
            self.get_faces()
        if include_boundary_faces:
            return self.F
        else:
            return self.F - len(self.get_boundary_faces())

    def is_planar_embedding(self):
        if self.face_edges is None:
            self.get_faces()
        return self.node_count() + self.face_count(False) == self.edge_count() + 1

    def get_areas(self, include_boundary_faces=True):
        if self.areas is None:
            if self.determinant is None:
                self._compute_determinant()
            self.areas = 0.5 * self.faces_v_array.sum(self.determinant)
        if self.boundary_face_indices is None:
            self.boundary_face_indices = np.flatnonzero((self.areas < 0) | np.isclose(self.areas, 0.0))
        if include_boundary_faces:
            return self.areas
        else:
            return np.delete(self.areas, self.boundary_face_indices)

    def get_centroids(self, include_boundary_faces=True):
        six_times_area = 6.0 * self.get_areas()
        mask = np.isclose(six_times_area, 0.0)
        six_times_area[mask] = 2.0 * self.face_lengths[mask]
        long_mask = self.faces_v_array.at_out_index(mask)
        X = self.x[self.face_nodes] + self.x[self.roll_ids]
        Y = self.y[self.face_nodes] + self.y[self.roll_ids]
        centroid_x = self.faces_v_array.sum(np.where(long_mask, X, self.determinant * X)) / six_times_area
        centroid_y = self.faces_v_array.sum(np.where(long_mask, Y, self.determinant * Y)) / six_times_area
        if include_boundary_faces:
            return centroid_x, centroid_y
        else:
            if self.boundary_face_indices is None:
                self.get_boundary_faces()
            return np.delete(centroid_x, self.boundary_face_indices), \
                   np.delete(centroid_y, self.boundary_face_indices)

    def get_num_components(self):
        if self.areas is None:
            self.get_areas()
        mask = np.ones(self.node_count(), dtype=bool)
        mask[self.node1] = False
        mask[self.node2] = False
        return len(self.boundary_face_indices) + np.sum(mask)

    def get_boundary_faces(self):
        if self.areas is None:
            self.get_areas()
        return self.boundary_face_indices

    def cycle_space(self, include_boundary_faces=False):
        face_edges, face_lengths = self.get_faces(include_boundary_faces=include_boundary_faces)
        E, F = self.edge_count(), len(face_lengths)
        indptr = np.append([0], np.cumsum(face_lengths))
        indices, data = face_edges % E, 1 - 2 * (face_edges // E)
        return scipy.sparse.csr_matrix((data, indices, indptr), shape=(F, E)).tocsc()

    def cut_space(self):
        E, N = self.edge_count(), self.node_count()
        row = np.concatenate((self.node1, self.node2))
        col = np.concatenate((np.arange(E), np.arange(E)))
        data = np.concatenate((-np.ones(E), np.ones(E)))
        return scipy.sparse.coo_matrix((data, (row, col)), shape=(N, E)).tocsc()

    def adjacency_matrix(self):
        E, N = self.edge_count(), self.node_count()
        data = np.ones(2 * E)
        row = np.append(self.node1, self.node2)
        col = np.append(self.node2, self.node1)
        return scipy.sparse.coo_matrix((data, (row, col)), shape=(N, N)).tocsc()

    def get_common_edge_of_faces(self, face1, face2, return_orientation=False):
        # returns index of an edge occurring in both faces if it exists; otherwise -1.
        # If multiple exist; returns the lowest index.
        # ({s},) ({s},) -> ({s},) in range(E) or -1
        # optionally return orientation; True if face1 passes edge counterclockwise
        # first encountering its node with lowest index
        if self.face_edges is None:
            self.get_faces()
        f = self.faces_v_array.rows()[np.argsort(self.face_edges)].astype(np.int64)
        E = self.edge_count()
        f1, f2 = f[:E], f[E:2 * E]
        mask = f1 < f2
        f1, f2 = np.where(mask, f1, f2),  np.where(mask, f2, f1)

        sorter = np.lexsort((f2, f1))
        f1, f2 = f1[sorter],  f2[sorter]
        A = np.array(list(zip(f1, f2)), dtype=[('f1', 'int'), ('f2', 'int')])

        face1, face2 = np.array(face1, dtype=int), np.array(face2, dtype=int)
        input_shape = face1.shape
        face1, face2 = face1.ravel(), face2.ravel()
        mask2 = face1 < face2
        fmin, fmax = np.where(mask2, face1, face2).astype(np.int64), np.where(mask2, face2, face1).astype(np.int64)
        V = np.array(list(zip(fmin, fmax)), dtype=[('fmin', 'int'), ('fmax', 'int')])

        edge_ids = np.searchsorted(A, V)
        edge_ids[edge_ids >= E] = E - 1
        if return_orientation:
            orientation = mask[sorter][edge_ids] ^ ~mask2

        found_mask = (f1[edge_ids] != fmin) | (f2[edge_ids] != fmax)
        edge_ids = sorter[edge_ids]

        edge_ids[found_mask] = -1

        if return_orientation:
            orientation[found_mask] = False
            return edge_ids.reshape(input_shape), orientation.reshape(input_shape)
        else:
            return edge_ids.reshape(input_shape)

    def permute_nodes(self, permutation):
        # first precompute faces, because face order will be different if computed after node permutation.
        if self.face_edges is None:
            self.get_faces()
        if self.face_nodes is None:
            self.get_face_nodes()
        if self.areas is None:
            self.get_areas()
        permutation = np.array(permutation, dtype=int)
        if not np.all(np.sort(permutation) == np.arange(self.node_count())):
            raise ValueError("invalid permutation")
        self.x = self.x[permutation]
        self.y = self.y[permutation]
        inv_perm = np.argsort(permutation)
        self.node1 = inv_perm[self.node1]
        self.node2 = inv_perm[self.node2]
        edge_sorter = self._sort_edges(self.node1, self.node2)
        self.edge_v_array = self._assign_edge_v_array()
        inv_edge_sorter = np.argsort(np.append(edge_sorter, (edge_sorter + self.edge_count()) % (2 * self.edge_count())))
        self.face_edges = inv_edge_sorter[self.face_edges]
        self.face_nodes = inv_perm[self.face_nodes]

    def permute_faces(self, permutation):
        permutation = np.array(permutation, dtype=int)
        if not np.all(np.sort(permutation) == np.arange(self.face_count())):
            raise ValueError("invalid permutation")
        inv_permutation = np.argsort(permutation)
        if self.face_edges is None:
            self.get_faces()
        self.face_lengths = self.face_lengths[permutation]
        self.face_edges = self.faces_v_array.permute_rows(permutation, self.face_edges)
        if self.areas is not None:
            self.areas = self.areas[permutation]
            self.determinant = self.faces_v_array.permute_rows(permutation, self.determinant)
            self.roll_ids = self.faces_v_array.permute_rows(permutation, self.roll_ids)
            self.boundary_face_indices = inv_permutation[self.boundary_face_indices]
        if self.face_nodes is not None:
            self.face_nodes = self.faces_v_array.permute_rows(permutation, self.face_nodes)
        self.faces_v_array = VarRowArray(self.faces_v_array.counts[permutation])

    def plot(self, show_faces=True, figsize=[5, 5], show_node_ids=False, show_edge_ids=False,
             show_face_ids=False, face_shrink_factor=0.9, show_boundary_face=False):
        self.fig, self.ax = plt.subplots(figsize=figsize)
        x, y = self.coo()
        self.ax.plot([x[self.node1], x[self.node2]], [y[self.node1], y[self.node2]], color=[0.5,0.5,0.5])
        self.ax.plot([x], [y], color=[0,0,0], marker="o", markerfacecolor=[0,0,0])
        if show_node_ids:
            for i, (xn, yn) in enumerate(zip(x, y)):
                self.ax.annotate(i.__str__(), (xn, yn))
        if show_edge_ids:
            x1, y1,  x2, y2 = x[self.node1], y[self.node1], x[self.node2], y[self.node2]
            for i, (xn, yn) in enumerate(zip(0.5 * (x1 + x2), 0.5 * (y1 + y2))):
                self.ax.annotate(i.__str__(), (xn, yn), color=[0.3, 0.5, 0.9], ha='center', va='center')
        if show_faces:
            xc, yc = self.get_centroids()
            pn = self.get_face_nodes(to_list=True)
            areas = self.get_areas()
            bi = self.boundary_face_indices
            nr=0
            for i, (xcn, ycn, n, area) in enumerate(zip(xc, yc, pn, areas)):
                if np.in1d(i, bi):
                    if show_boundary_face:
                        xp, yp = x[n], y[n]
                        self.ax.plot(np.append(xp, xp[0]), np.append(yp, yp[0]), color=[0.2, 0.5, 1])
                        if show_face_ids:
                            self.ax.annotate(nr.__str__(), (xcn, ycn), color=[0.2, 0.5, 1], ha='center', va='center')
                            nr+=1
                else:
                    xp = face_shrink_factor * x[n] + (1 - face_shrink_factor) * xcn
                    yp = face_shrink_factor * y[n] + (1 - face_shrink_factor) * ycn
                    self.ax.plot(np.append(xp, xp[0]), np.append(yp, yp[0]), color=[1, 0.5, 0.2])
                    if show_face_ids:
                        self.ax.annotate(nr.__str__(), (xcn, ycn), color=[1, 0.5, 0.2], ha='center', va='center')
                        nr += 1
        return self.fig, self.ax

    def _sort_edges(self, node1, node2):
        node1, node2 = np.array(node1, dtype=int),  np.array(node2, dtype=int)
        mask = node1 < node2
        self.node1, self.node2 = np.where(mask, node1, node2), np.where(mask, node2, node1)
        self.edge_sorter = np.lexsort((self.node2, self.node1))
        self.node1, self.node2 = self.node1[self.edge_sorter], self.node2[self.edge_sorter]
        return self.edge_sorter + (~mask).astype(int) * self.edge_count()

    def _compute_determinant(self):
        if self.roll_ids is None:
            self._compute_roll_ids()
        self.determinant = self.x[self.roll_ids] * self.y[self.face_nodes] - self.x[self.face_nodes] * self.y[self.roll_ids]

    def _compute_roll_ids(self):
        if self.face_edges is None:
            self.get_faces()
        if self.face_nodes is None:
            self.get_face_nodes()
        self.roll_ids = self.face_nodes[self.faces_v_array.roll(1)]

    def _get_edge_map(self):
        """
        Computes a one-to-one map over directed edges, where the image contains the edge
        which is encountered next when traversing the graph moving counter-clockwise.

        Used to generate faces by repeating indexing: e_(i+1) = map[e_i]

        map: np.arange(2*E) -> sorted_out_edge_directed

        Output:
        sorted_out_edge_directed    (2*E,) in range(2*E)
        """

        counter_clockwise = True
        edge_count = self.edge_count()
        ns = np.append(self.node1, self.node2)

        # construct neighbour structure
        nodes, count = np.unique(ns, return_counts=True)
        neighbours = VarRowArray(count)
        neighbour_edges = np.tile(np.arange(edge_count), 2)[np.argsort(ns)]
        neighbour_node_self = nodes[neighbours.rows()]
        neighbour_node_other = np.where(self.node1[neighbour_edges] == neighbour_node_self,
                                        self.node2[neighbour_edges], self.node1[neighbour_edges])

        # sort neighbour in-dimension by ascending angle
        angles = np.arctan2(self.y[neighbour_node_other] - self.y[neighbour_node_self],
                            self.x[neighbour_node_other] - self.x[neighbour_node_self])
        angle_arg_sort = np.argsort(3 * np.pi * neighbour_node_self.astype(np.double) + angles)
        neighbour_edges = neighbour_edges[angle_arg_sort]
        neighbour_node_other = neighbour_node_other[angle_arg_sort]

        def to_directed(edge_nr, edge_direction, edge_count):
            return edge_nr + edge_count * (1 - edge_direction.astype(int))

        # create combined-index for input edges of the map
        neighbour_edges_direction = neighbour_node_other < neighbour_node_self
        in_edge_combined_index = to_directed(neighbour_edges, neighbour_edges_direction, edge_count)

        # find the map from every (combined index) edge to the next edge in (counter-)-clockwise direction
        edges_map = neighbours.roll() if counter_clockwise else neighbours.roll(-1)

        # find the combined index of the output edges of the map
        out_edge_directed = to_directed(neighbour_edges, ~neighbour_edges_direction, edge_count)[edges_map]

        # sort the edge-map based on the (combined index of the) input edges
        sorted_out_edge_directed = out_edge_directed[np.argsort(in_edge_combined_index)]

        return sorted_out_edge_directed

    def _construct_faces(self):
        """
        Computes faces of the embedded graph. This is done using repeated iteration over the
        one-to-one map over directed edges computed with _get_edge_map(). It starts with all
        edges pointing from lowest to highest node to ensure all faces are found.

        Used to generate faces by repeating indexing: e_(i+1) = map[e_i]

        map: np.arange(2*E) -> sorted_out_edge_directed

        Output:
        sorted_out_edge_directed    (2*E,) in range(2*E)
        """

        map = self._get_edge_map()
        edge_ids = np.arange(self.edge_count())
        paths = -np.ones((3, self.edge_count()), dtype=int)
        paths[0, :] = edge_ids
        path_lengths = np.ones(self.edge_count(), dtype=int)
        current_path_length, out_paths, out_path_lengths = 1, np.zeros(0, dtype=int), np.zeros(0, dtype=int)

        # iteration doing counter-clockwise walks through the graphs starting from each junction
        while len(edge_ids) > 0:
            edge_ids = map[edge_ids]
            is_terminated = edge_ids == paths[0, :]
            if np.any(is_terminated):
                out_path_lengths, out_paths = self._store_terminated_paths(paths[:current_path_length, is_terminated].T,
                                                                            out_path_lengths, out_paths)
                paths, path_lengths = paths[:, ~is_terminated], path_lengths[~is_terminated]
                edge_ids = edge_ids[~is_terminated]
            paths[current_path_length, :] = edge_ids
            current_path_length += 1
            if paths.shape[0] == (current_path_length):
                paths = np.append(paths, -np.ones(paths.shape, dtype=int), axis=0)
        return out_paths, out_path_lengths

    def _store_terminated_paths(self, terminated_paths, out_path_lengths, out_paths):
        # roll paths until the lowest node index is in the first column (so its easier to remove duplicate paths)
        path_cnt, path_len = terminated_paths.shape
        paths = terminated_paths[np.arange(path_cnt)[:, None], (
                    np.arange(path_len) + np.argmin(terminated_paths, axis=1)[:, None]) % path_len]

        _, idx = np.unique(paths[:, 0], return_index=True)
        paths = paths[idx, :]
        #
        # paths = np.unique(paths, axis=0)

        out_path_lengths = np.append(out_path_lengths, path_len * np.ones(paths.shape[0], dtype=int))
        out_paths = np.append(out_paths, paths.flatten())
        return out_path_lengths, out_paths

    def _non_boundary_mask(self):
        out = np.ones(self.face_count(include_boundary_faces=True), dtype=bool)
        out[self.boundary_face_indices] = False
        return out

    def _cycle_space_solve_for_integral_x(self, b):
        """
        Solves the equation: A @ x = b (where A = cycle_matrix, without boundary faces).
        If b is integral (contain only integers), the output array x will also be integral.

        input:  b (..., F)
        output: x (..., E)

        Notes:
            - The equation is underdetermined, so the solution x is not unique.
        """
        if (self.get_num_components() != 1) or (not self.is_planar_embedding()):
            raise ValueError("only implemented for single component planar embedding")

        E, F = self.edge_count(), self.face_count(include_boundary_faces=False)
        b = np.array(b, dtype=np.double)
        b_shape = list(b.shape)
        b = b.reshape(-1, F)
        b_tally = b.copy()

        # insert boundary face in b_tally
        b_idx = self.boundary_face_indices[0]
        b_tally = np.concatenate((b_tally[:, :b_idx], np.zeros((b_tally.shape[0], 1)), b_tally[:, b_idx:]), axis=1)

        # do depth first search, resulting in cur (current node of tree) and prev (parent node of cur)
        A = self.cycle_space(include_boundary_faces=True)
        cur, predecessor = scipy.sparse.csgraph.depth_first_order(A @ A.T, b_idx)
        prev = predecessor[cur]
        prev[0] = -1

        # find map of edge between pair of faces. signs returns +1 if face1 counterclockwise passes resulting edge in its own direction.
        juncs, orientation_in_face_1 = self.get_common_edge_of_faces(cur, prev, return_orientation=True)
        sgns = (-1 + 2 * orientation_in_face_1.astype(np.double))

        # construct x at each edge by passing through the tree in reverse.
        x = np.zeros((b.shape[0], E), dtype=b.dtype)
        for i in reversed(range(F+1)):
            if prev[i] >= 0:
                b_tally[:, prev[i]] += b_tally[:, cur[i]]
                x[:, juncs[i]] += b_tally[:, cur[i]] * sgns[i]
                b_tally[:, cur[i]] = 0

        # check if resulting x solves A @ x == b
        b_shape[-1] = E
        mask = self._non_boundary_mask()
        if not np.allclose((A @ x.T)[mask, :], b.T):
            raise ValueError("failed integral solve of cycle space linear problem")

        # return x
        return x.reshape(tuple(b_shape))

class EmbeddedSquareGraph(EmbeddedGraph):

    def __init__(self, count_x, count_y,  x_scale=1.0, y_scale=1.0):
        y, x = np.mgrid[0:count_y, 0:count_x]
        idx = np.arange(count_x * count_y).reshape(count_y, count_x)
        n1 = np.concatenate((idx[:, 0:-1].ravel(), idx[0:-1, :].ravel()))
        n2 = np.concatenate((idx[:, 1:].ravel(), idx[1:, :].ravel()))
        super().__init__(x * x_scale, y * y_scale, n1, n2)


class EmbeddedHoneycombGraph(EmbeddedGraph):

    def __init__(self, count_x, count_y, x_scale=1.0, y_scale=1.0):
        y, x = np.mgrid[0:count_y, 0:count_x]
        x1, y1 = 3.0 * x, np.sqrt(3.0) * y
        nodes_x = np.concatenate((x1, x1 + 0.5, x1 + 1.5, x1 + 2), axis=0).ravel()
        nodes_y = np.concatenate((y1, y1 + np.sqrt(0.75), y1 + np.sqrt(0.75), y1), axis=0).ravel()
        idx = np.arange(count_x * count_y).reshape(count_y, count_x)
        s = count_x * count_y
        nodes1 = (idx,   idx[:-1, :]+s, idx+s,   idx[:-1, :]+2*s, idx+2*s, idx[:, :-1]+3*s)
        nodes2 = (idx+s, idx[1:, :],    idx+2*s, idx[1:, :]+3*s,  idx+3*s, idx[:, 1:])
        nodes1 = np.concatenate(tuple([n1.flatten() for n1 in nodes1])).ravel()
        nodes2 = np.concatenate(tuple([n2.flatten() for n2 in nodes2])).ravel()

        remove_node_ids = [0, idx[0, -1] + 3 * s]
        nodes_x, nodes_y, nodes1, nodes2 = remove_nodes(nodes_x, nodes_y, nodes1, nodes2, remove_node_ids)

        super().__init__(nodes_x * x_scale, nodes_y * y_scale, nodes1, nodes2)


class EmbeddedTriangularGraph(EmbeddedGraph):

    def __init__(self, count_x, count_y, x_scale=1.0, y_scale=1.0):
        y, x = np.mgrid[0:count_y, 0:count_x]
        x1, y1 = x, np.sqrt(3.0) * y
        nodes_x = np.concatenate((x1, x1 + 0.5), axis=0)
        nodes_y = np.concatenate((y1, y1 + np.sqrt(0.75)), axis=0)
        idx = np.arange(count_x * count_y).reshape(count_y, count_x)
        s = count_x * count_y
        nodes1 = (idx,   idx[:, :-1], idx[:-1, :]+s, idx[:-1, :-1]+s, idx[:, :-1]+s, idx[:, 1:])
        nodes2 = (idx+s, idx[:, 1:],  idx[1:, :],    idx[1:, 1:],     idx[:, 1:]+s,  idx[:, :-1]+s)
        nodes1 = np.concatenate(tuple([n1.flatten() for n1 in nodes1]))
        nodes2 = np.concatenate(tuple([n2.flatten() for n2 in nodes2]))
        super().__init__(nodes_x * x_scale, nodes_y * y_scale, nodes1, nodes2)

def remove_nodes(x, y, nodes1, nodes2, remove_node_ids):
    remove_node_ids = np.array(remove_node_ids)
    node_map = np.ones(len(x), dtype=int)
    node_map[remove_node_ids] = 0
    node_map = np.cumsum(node_map) - node_map
    remove_edges = np.in1d(nodes1, remove_node_ids) | np.in1d(nodes2, remove_node_ids)
    x = np.delete(x, remove_node_ids)
    y = np.delete(y, remove_node_ids)
    nodes1 = node_map[nodes1[~remove_edges]]
    nodes2 = node_map[nodes2[~remove_edges]]
    return x, y, nodes1, nodes2

def remove_edges(nodes1, nodes2, remove_edge_ids):
    remove_edge_ids = np.array(remove_edge_ids)
    edge_map = np.ones(len(nodes1), dtype=int)
    edge_map[remove_edge_ids] = 0
    edge_map = np.cumsum(edge_map) - edge_map
    nodes1 = edge_map[np.delete(nodes1, remove_edge_ids)]
    nodes2 = edge_map[np.delete(nodes2, remove_edge_ids)]
    return nodes1, nodes2



import numpy as np

from pyJJAsim import *

import matplotlib
matplotlib.use("TkAgg")
import matplotlib.pyplot as plt


"""
EXAMPLE 4: Parameter optimization

Compute maximal frustration bounds and maximal current 
"""

if __name__ == "__main__":
    # define arrays
    array = SquareArray(12, 12)


    # compute frustration bounds for zero vortex state
    prob = StaticProblem(array, current_sources=0)
    (smallest_f, largest_f), (s_config, l_config), _ = prob.compute_frustration_bounds()
    print(f"smallest and largest frustration for which the zero-vortex state exists: {smallest_f}, {largest_f}")
    s_config.plot(title=f"minimal frustration in zero vortex state (f={smallest_f})")

    # compute maximal current
    prob = StaticProblem(array, frustration=0, current_sources=array.vertical_junctions())
    I_factor, net_I, max_I_config, _ = prob.compute_maximal_current()
    np.set_printoptions(linewidth=10000000)
    print(f"largest current factor {I_factor} (corresponding to net current of  {net_I}) at which the zero-vortex state exists at zero frustration")
    max_I_config.plot(title=f"maximal current in zero vortex state (net_I={net_I})")

    # compute extermum in Is-f space using compute_stable_region()
    prob = StaticProblem(array, frustration=0, current_sources=array.vertical_junctions())
    f, net_I, _, _ = prob.compute_stable_region()
    plt.subplots()
    plt.plot(f, net_I)
    plt.xlabel("frustration")
    plt.ylabel("net I")
    plt.title("Region in f-I parameter space where the zero-vortex state is stable")

    # compute direction dependent maximal current using compute_maximal_parameter()
    prob = StaticProblem(array)
    angles = np.linspace(0, 2*np.pi, 33)
    Ih, Iv = array.horizontal_junctions(), array.vertical_junctions()
    I_factor = np.zeros(len(angles))
    f_func = lambda x: 0
    for i, angle in enumerate(angles):
        Is_func = lambda x: x * (np.cos(angle) * Ih + np.sin(angle) * Iv)
        I_factor[i], _, _, _ = prob.compute_maximal_parameter(Is_func, f_func)
    plt.subplots()
    plt.plot(np.cos(angles) * I_factor, np.sin(angles) * I_factor, marker="o")
    plt.xlabel("horizontal current")
    plt.ylabel("vertical current")
    plt.title("angle dependent maximal current for zero-vortex state")
    plt.show()



from __future__ import annotations

import time

import numpy as np
import scipy
import scipy.sparse.linalg
import scipy.optimize
from scipy.sparse.linalg import ArpackNoConvergence

from pyJJAsim.compute import Matrix
from pyJJAsim.josephson_circuit import Circuit

__all__ = ["CurrentPhaseRelation", "DefaultCPR",
           "StaticProblem", "StaticConfiguration", "node_to_junction_current"]

DEF_TOL = 1E-11

DEF_NEWTON_MAXITER = 30
DEF_STAB_MAXITER = 100

DEF_MAX_PAR_TOL = 1E-4
DEF_MAX_PAR_REDUCE_FACT = 0.42

class CurrentPhaseRelation:

    """

    Current-Phase relation Icp(Ic, theta). The default value is Icp = Ic * sin(theta).

    Input:
    ------
    func            lambda Ic, theta: ...   current-phase relation
    d_func          lambda Ic, theta: ...   derivative of current-phase relation to theta
    i_func          lambda Ic, theta: ...   integral of current-phase relation over theta (starting at 0)

    Methods:
    --------
    eval(Ic: np.ndarray, theta: np.ndarray) -> np.ndarray
    d_eval(Ic: np.ndarray, theta: np.ndarray) -> np.ndarray
    i_eval(Ic: np.ndarray, theta: np.ndarray) -> np.ndarray

    Remarks:
    --------
     - func, d_func and i_func must be numpy ufunc, so their output must be broadcast of input Ic and theta.
    """
    def __init__(self, func, d_func, i_func):
        self.func = func
        self.d_func = d_func
        self.i_func = i_func

    def eval(self, Ic, theta):
        return self.func(Ic, theta)

    def d_eval(self, Ic, theta):
        return self.d_func(Ic, theta)

    def i_eval(self, Ic, theta):
        return self.i_func(Ic, theta)

class DefaultCPR(CurrentPhaseRelation):

    """
    Default current-phase relation Icp = Ic * sin(theta).

    """
    def __init__(self):
        super().__init__(lambda Ic, th: Ic * np.sin(th),
                         lambda Ic, th: Ic * np.cos(th),
                         lambda Ic, th: Ic * (1.0 - np.cos(th)))


class NewtonIterInfo:

    """
    Information about the newton iteration used to find static configurations.
    Use print(newton_iter_info) to display the information.

    Methods:
    -------
    get_max_iter()                          Returns number of iterations after which iteration is aborted.
    get_tol()                               Returns tolerance.
    get_status()                            Returns 0: converged. residual < tolerance
                                                    1: diverged before reaching maxiter.
                                                    2: reached max_iter without converging or diverging.
    has_converged()                         Returns if iteration has converged.
    get_is_target_vortex_configuration()    Returns (nr_of_iters,) bool array if vortex configuration at iter
                                            agrees with vortex configuration specified in problem.
    found_target_solution()                 Returns True if has_converged() and final iter obeys target vortex config.
    get_number_of_iterations()              Returns number of newton iterations done.
    get_residual()                          (nr_of_iters,) array containing residual at each iteration.
    get_runtime()                           Returns runtime in seconds.
    plot_residuals()                        Plots residual vs iteration number.

    """
    def __init__(self, tol, maxiter):
        self.start_time = time.perf_counter()
        self.tol = tol
        self.maxiter = maxiter
        self.iteration = 0
        self.error = np.zeros(self.maxiter + 1, dtype=np.double)
        self.has_converged = False
        self.is_target_n = np.zeros(self.maxiter + 1, dtype=int)
        self.runtime = 0.0

    def get_max_iter(self):
        return self.maxiter

    def get_tol(self):
        return self.tol

    def get_status(self):
        return int(not self.found_target_solution()) + 2 * int(self.iteration >= self.maxiter)

    def has_converged(self):
        return self.has_converged

    def get_is_target_vortex_configuration(self):
        return self.is_target_n[:(self.get_number_of_iterations()+1)]

    def found_target_solution(self):
        return self.has_converged and self.is_target_n[self._get_iteration()]

    def get_number_of_iterations(self):
        return self._get_iteration()

    def get_residual(self):
        return self.error[:(self.get_number_of_iterations()+1)]

    def get_runtime(self):
        return self.runtime

    def plot_residuals(self):
        import matplotlib.pyplot as plt
        n = self.get_is_target_vortex_configuration().astype(bool)
        y = self.get_residual()
        x = np.arange(len(y))
        plt.semilogy(x[n], y[n], color=[0, 0, 0], label="n is target_n", linestyle="None", marker="o")
        plt.semilogy(x[~n], y[~n], color=[1, 0, 0], label="n is not target_n", linestyle="None", marker="o")
        plt.xlabel("Newton iteration number")
        plt.ylabel("residual")
        plt.title("Evolution of residual for newton iteration.")
        plt.legend()

    def __str__(self):
        out = f"newton iteration info: (tol={ self.get_tol()}, maxiter={self.get_max_iter()})\n\t"
        out += f"status: {self.get_status()}"
        if self.get_status() == 0:
            out += f" (converged)\n\t"
        if self.get_status() == 1:
            out += f" (diverged)\n\t"
        if self.get_status() == 2:
            out += f" (indeterminate; reached max_iter without converging or diverging)\n\t"
        out += f"number of iterations: {self.get_number_of_iterations()}\n\t"
        out += f"residual: {self.get_residual()}\n\t"
        out += f"runtime (in sec): {self.get_runtime()}"
        return out

    def _set(self, error, is_target_n_v):
        self.has_converged = error < self.tol
        self.is_target_n[self.iteration] = is_target_n_v
        self.error[self.iteration] = error
        self.runtime += time.perf_counter() - self.start_time
        self.start_time = time.perf_counter()
        self.iteration += 1
        return self

    def _get_iteration(self):
        return max(0, self.iteration - 1)


class ParameterOptimizeInfo:

    """
    Information about the parameter optimization process.
    Use print(parameter_optimize_info) to display the information.

    Methods:
    -------
    get_has_stable_target_solution_at_zero()    Returns if a stable target solution is found at lambda=0
    get_lambda()                                Returns (nr_of_steps,) array with lambda at each step
    get_lambda_error()                          Returns (nr_of_steps,) array with error in lambda
    get_lambda_lower_bound()                    Returns lower bound for lambda
    get_lambda_upper_bound()                    Returns upper bound for lambda
    get_found_stable_target_solution()          Returns (nr_of_steps,) array if a stable target solution is found at step
    get_newton_converged()                      Returns (nr_of_steps,) array if newton iteration converged at step
    get_newton_target_n()                       Returns (nr_of_steps,) array if target n at step
    get_newton_steps()                          Returns (nr_of_steps,) array with nr of newton iterations at step
    get_newton_iter_all_info()                  Returns (nr_of_steps,) list containing newton_iter_infos
    get_is_stable()                             Returns (nr_of_steps,) array if a stable at step
    get_stability_steps()                       Returns (nr_of_steps,) array with stability algorithm iterations at step
    get_stable_iter_all_info()                  Returns (nr_of_steps,) list containing stable_iter_infos
    get_runtime()                               Returns runtime in seconds.
    plot_residuals()                            Plots residual vs iteration number.

    """
    def __init__(self, Is_func, f_func, lambda_tol, M):
        self.Is_func = Is_func
        self.f_func = f_func
        self.lambda_tol = lambda_tol
        self.has_stable_target_solution_at_zero = False
        self.lambda_history = np.zeros(1000, dtype=np.double)
        self.stepsize_history = np.zeros(1000, dtype=np.double)
        self.solution_history = np.zeros(1000, dtype=np.bool)
        self.newton_iter_infos = []
        self.M = M
        self._step = 0
        self._time = time.perf_counter()

    def get_has_stable_target_solution_at_zero(self):
        return self.has_stable_target_solution_at_zero

    def get_lambda(self):
        return self.lambda_history[:self._step]

    def get_lambda_error(self):
        return self._get_lambda_stepsize() / self.get_lambda()

    def get_lambda_lower_bound(self):
        if not self.get_has_stable_target_solution_at_zero():
            return np.nan
        s = self.get_lambda()[self.get_found_stable_target_solution()]
        return s[-1] if s.size > 0 else 0

    def get_lambda_upper_bound(self):
        s = self.get_lambda()[~self.get_found_stable_target_solution()]
        return s[-1] if s.size > 0 else np.inf

    def get_found_stable_target_solution(self):
        return self.solution_history[:self._step]

    def get_newton_iter_all_info(self):
        return self.newton_iter_infos

    def get_newton_converged(self):
        return np.array([info.converged() for info in self.newton_iter_infos], dtype=int)

    def get_newton_target_n(self):
        return np.array([info.get_is_target_vortex_configuration()[-1] for info in self.newton_iter_infos], dtype=int)

    def get_newton_steps(self):
        return np.array([info.get_number_of_iterations() for info in self.newton_iter_infos], dtype=int)

    def get_runtime(self):
        return self._time

    def plot_residuals(self):
        import matplotlib.pyplot as plt
        for i, n_info in enumerate(self.get_newton_iter_all_info()):
            n = n_info.get_is_target_vortex_configuration().astype(bool)
            y = n_info.get_residual()
            x = np.arange(len(y))
            plt.semilogy(x[n], y[n], color=[0, 0, 0], linestyle="None", marker="o")
            plt.semilogy(x[~n], y[~n], color=[1, 0, 0], linestyle="None", marker="o")
            plt.text(x[-1], y[-1], str(i), color=[0.3, 0.3, 0.3])
            plt.semilogy(x, y, color=[0, 0, 0])
        plt.xlabel("Newton iteration number")
        plt.ylabel("residual")
        plt.title("Newton residuals in parameter optimization.")
        plt.legend(["n is target_n", "n is not target_n"])

    def __str__(self):
        np.set_printoptions(linewidth=100000)
        out = "parameter optimize info:\n\t"
        if not self.get_has_stable_target_solution_at_zero():
            out += "optimization failed because not solution was found at lambda=0"
        else:
            def int_digit_count(x):
                return np.ceil(np.log(np.max(x)) / np.log(10)).astype(int)
            n = max(5, 3 + int_digit_count(1/self.lambda_tol), int_digit_count(self.get_newton_steps()))
            out += f"Found lambda between {self.get_lambda_lower_bound()} and {self.get_lambda_upper_bound()}"
            if self.last_step_status != 2:
                out += f" at desired tolerance (resid={self.get_lambda_error()[-1]}) \n\t"
            else:
                out += f"; stopped because newton iteration was indeterminate. Consider increasing newton_maxiter.)\n\t"
            out += f"runtime: {np.round(self.get_runtime(), 7)} sec\n\t"
            out += f"lambda: {self.get_lambda()}\n\t"
            out += f"net I:  {self._get_net_I()}\n\t"
            out += f"mean f: {self._get_mean_f()}\n\t"
            np.set_printoptions(formatter={'float': lambda x: ("{0:0." + str(n - 2) + "f}").format(x)})
            out += f"lambda / lambda[0]:    {self.get_lambda() / self.get_lambda()[0]}\n\t"
            np.set_printoptions(formatter={'bool': lambda x: ("{:>" + str(n) + "}").format(x)})
            out += f"newton converged:      {self.get_found_stable_target_solution().astype(bool)}\n\t"
            out += f"to target n:           {self.get_newton_target_n().astype(bool)}\n\t"
            np.set_printoptions(formatter={'int': lambda x: ("{:>" + str(n) + "}").format(x)})
            out += f"total newton steps:    {self.get_newton_steps()}\n\t"
        return out

    def _preset(self, has_stable_target_solution_at_zero):
        self.has_stable_target_solution_at_zero = has_stable_target_solution_at_zero
        return self

    def _set(self, lambda_value, lambda_stepsize, found_stable_target_solution, newton_iter_info):
        self.lambda_history[self._step] = lambda_value
        self.stepsize_history[self._step] = lambda_stepsize
        self.solution_history[self._step] = found_stable_target_solution
        self.newton_iter_infos += [newton_iter_info]
        self._step += 1
        return self

    def _finish(self, last_step_status):
        self.last_step_status = last_step_status
        self._time = time.perf_counter() - self._time
        return self

    def _get_lambda_stepsize(self):
        return self.stepsize_history[:self._step]

    def _get_net_I(self):
        return np.array([0.5 * np.sum(np.abs(self.M @ np.broadcast_to(self.Is_func(l), (self.M.shape[1],)))) for l in self.get_lambda()])

    def _get_mean_f(self):
        return np.array([np.mean(self.f_func(l)) for l in self.get_lambda()])


class StaticProblem:
    """
    Define a static josephson junction array problem.

    Input:
    ------                  symbol  type                      getter
    circuit                         Circuit                   get_circuit()
    current_sources         Is      (Nj,) ndarray or scalar   get_current_sources()
    frustration             f       (Nf,) ndarray or scalar   get_frustration()
    vortex_configuration    n       (Nf,) ndarray or scalar   get_vortex_configuration()
    current_phase_relation  cp      CurrentPhaseRelation      get_current_phase_relation()

    Methods:
    --------
    get_net_sourced_current         Gets the sum of all (positive) current injected at nodes to create Is.
    new_problem                     Makes copy of self with specified modifications.
    approximate                     Computes approximate solution (London approximation or arctan approximation)
    approximate_placed_vortices     Computes arctan approximation where vortices are placed at specified
                                    coordinates, rather than in face centers.
    compute                         Compute exact solution
    compute_maximal_parameter       Computes largest x where stable solution exists with f=f(x) and Is=Is(x)
    compute_frustration_bounds      Compute smallest and largest frustration with a stable solution
    compute_maximal_current         Compute largest x where Is=x * self.Is with a stable solution
    compute_stable_region           Compute boundary of stable region in f-Is space of vortex configuration n.

    get_net_sourced_current() -> scalar

    new_problem(current_sources=None, frustration=None, vortex_configuration=None,
                current_phase_relation=None) -> StaticProblem

    approximate(algorithm=1) -> StaticConfiguration

    approximate_placed_vortices(n, x_n, y_n) -> StaticConfiguration

    compute(initial_guess: StaticConfiguration, ...) -> StaticConfiguration, int, NewtonIterInfo

    compute_maximal_parameter(Is_func, f_func, initial_guess: StaticConfiguration, ...)
     -> lower_bound, upper_bound, StaticConfiguration, ParameterOptimizeInfo

    compute_frustration_bounds(initial_guess: StaticConfiguration, start_frustration, ...)
     -> (smallest_f, largest_f), (StaticConfiguration, StaticConfiguration), (ParameterOptimizeInfo, ParameterOptimizeInfo)

    compute_maximal_current(initial_guess: StaticConfiguration, ...)
     -> Is_factor, net_I, StaticConfiguration, ParameterOptimizeInfo

    compute_stable_region(num_angles=40, start_frustration, start_initial_guess, ...)
     -> [f], [net_I], [StaticConfiguration], [ParameterOptimizeInfo]

    """

    def __init__(self, circuit: Circuit, current_sources=0.0, frustration=0.0,
                 vortex_configuration=0, current_phase_relation=DefaultCPR()):
        self.circuit = circuit
        self.current_sources = np.atleast_1d(current_sources)
        self.frustration = np.atleast_1d(frustration)
        self.vortex_configuration = np.atleast_1d(vortex_configuration)
        self.current_phase_relation = current_phase_relation
        self.current_sources_norm = None
        self.Asq_factorization = None
        self.AIpLIcA_factorization = None
        self.IpLIc_factorization = None
        self.Msq_factorization = None

    def get_circuit(self) -> Circuit:
        return self.circuit

    def get_current_sources(self):
        return self.current_sources

    def get_frustration(self):
        return self.frustration

    def get_vortex_configuration(self):
        return self.vortex_configuration

    def get_current_phase_relation(self):
        return self.current_phase_relation

    def new_problem(self, current_sources=None, frustration=None,
                    vortex_configuration=None, current_phase_relation=None):
        """
        Makes copy of self with specified modifications.
        """
        return StaticProblem(self.circuit, current_sources=self.current_sources if current_sources is None else current_sources,
                             frustration=self.frustration if frustration is None else frustration,
                             vortex_configuration=self.vortex_configuration if vortex_configuration is None else vortex_configuration,
                             current_phase_relation=self.current_phase_relation if current_phase_relation is None else current_phase_relation)

    def get_phase_zone(self):
        return 0

    def get_net_sourced_current(self):
        """
        Gets the sum of all (positive) current injected at nodes to create Is.
        """
        M = self.get_circuit().get_cut_matrix()
        return 0.5 * np.sum(np.abs((M @ self._Is())), axis=0)

    def approximate(self, algorithm=1) -> StaticConfiguration:
        """
        Computes approximate solutions.  Has two algorithms; the arctan approximation or
        the London approximation.

        alg: name                  description
         0   arctan approximation  assigns phases that "wind" 2*pi around vortices in z=0 phase zone,
                                   phi(x,y) = sum_i 2 * pi * n_i * atan2(y-y_n_i,x-x_n_i)
                                   where vortices are located at centres of their respective faces.
         1   london approximation  Find theta in cycle space (theta = A.T @ ...) that obeys winding rule.
        """
        if algorithm == 0:
            theta = arctan_approximation(self.circuit, self._f(), self._nt(),
                                         Asq_solver=self._Asq_factorization(),
                                         IpLIc_solver=self._IpLIc_factorization())
        elif algorithm == 1:
            theta = london_approximation(self.circuit, self._f(), self._nt(),
                                         AIpLIcA_solver=self._AIpLIcA_factorization())
            theta = change_phase_zone(self.get_circuit(), theta, self._nt(), 0)
        else:
            raise ValueError("invalid algorithm")
        return StaticConfiguration(self, theta)

    def approximate_placed_vortices(self, n, x_n, y_n) -> StaticConfiguration:
        """
        Compute arctan approximation with manual placement of vortices.

        Input:
        ------
        n       (N,) int array      vorticity at location (x_n, y_n)
        x_n     (N,) float array    x-coordinates of vortices
        y_n     (N,) float array    y-coordinates of vortices
        """

        theta = arctan_approximation_placed_vortices(self.circuit,
            self._f(), n, x_n, y_n, Asq_solver=self._Asq_factorization(), IpLIc_solver=self._IpLIc_factorization())

        return StaticConfiguration(self, theta)


    def compute(self, initial_guess: StaticConfiguration | np.ndarray = None,
                tol=DEF_TOL, maxiter=DEF_NEWTON_MAXITER, stop_as_residual_increases=True,
                stop_if_not_target_n=False) -> tuple[StaticConfiguration, int, NewtonIterInfo]:

        """
        Compute solutions using Newton iteration.

        Input:
        ------
        initial_guess=None                  None (London approximation is used)
                                            or (Nj,) np.ndarray representing theta
                                            or StaticConfiguration
        tol=DEF_TOL                         scalar      tolerance; is solution if |residual| < tol
        maxiter=DEF_NEWTON_MAXITER          int         maximum number of newton iterations.
        stop_if_not_target_n=False          bool        iteration stops  if n(iter) != n (diverged)
        stop_as_residual_increases=True     bool        iteration stops if error(iter) > error(iter - 3)  (diverged)

        Output:
        -------
        config      StaticConfiguration     object containing solution
        status      int                     0       converged
                                            or 1    diverged if error(iter)>0.5 or above reasons
                                            or 2    max_iter reached without converging or diverging
        iter_info   NewtonIterInfo          handle containing information about newton iteration
        """
        if initial_guess is None:
            initial_guess = self.approximate(algorithm=1)

        if isinstance(initial_guess, StaticConfiguration):
            initial_guess = initial_guess._th()

        theta, status, iter_info = static_compute(self.get_circuit(), initial_guess, Is=self._Is(),
                                                  f=self._f(), n=self._nt(), z=0,
                                                  cp=self.current_phase_relation, tol=tol,
                                                  maxiter=maxiter, Asq_solver=self._Asq_factorization(),
                                                  stop_as_residual_increases=stop_as_residual_increases,
                                                  stop_if_not_target_n=stop_if_not_target_n)
        config = StaticConfiguration(self, theta)
        return config, status, iter_info

    def compute_maximal_parameter(self, Is_func, f_func,
                                  initial_guess: StaticConfiguration | np.ndarray = None,
                                  lambda_tol=DEF_MAX_PAR_TOL, estimated_upper_bound=1.0,
                                  newton_tol=DEF_TOL, newton_maxiter=DEF_NEWTON_MAXITER,
                                  newton_stop_as_residual_increases=True,
                                  newton_stop_if_not_target_n=False,
                                  require_stability=True, stable_maxiter=DEF_STAB_MAXITER):

        """
        Finds the largest value of lambda for which a problem which current_sources=Is_func(lambda)
        and frustration=f_func(lambda) has a stable stationary state.

         - Must be able to find a stable configuration at lambda=0.
         - One can manually specify an initial_guess for lambda=0.
         - returns a lower- and upperbound for lambda. Stops when the difference < lambda_tol * lower_bound
         - furthermore returns config containing the solutions at the lower_bound. Also its
           accompanied problem has f and Is of lower_bound.
         - ignores self.current_sources and self.frustration.
         - Also returns ParameterOptimizeInfo object containing information about the iteration.
         - Algorithm stops if lambda_tol is reached or when newton_iteration failed to converge or diverge.
         - Algorithm needs an estimate of the upperbound for lambda to work.

        Input:
        ------
        Is_func                                 Is_func(lambda) -> Is
        f_func                                  f_func(lambda) -> f
        initial_guess=None                      manual approximation at lambda=0
                                                or None; then London approximation is used.
        lambda_tol=DEF_MAX_PAR_TOL              stop iterating if upperbound - lowerbound < lambda_tol * lower_bound
        estimated_upper_bound=1.0               Estimate for the upperbound for lambda

        (for newton iteration at given lambda, see .compute() for documentation)
        newton_tol=DEF_TOL
        max_total_newton_steps=DEF_MAX_PAR_NEWT_STEPS
        newton_stop_as_residual_increases=True
        newton_stop_if_not_target_n=False

        (for determining stability of solution at given lambda, see StaticConfiguration.is_stable())
        require_stability=True
        stable_maxiter=DEF_STAB_MAXITER

        Output:
        -------
        lambda_lowerbound       float                   Lowerbound of lambda
        lambda_upperbound       float                   Upperbound of lambda
        config                  StaticConfiguration     Containing solutions at lambda=lambda_lowerbound
        iteration_info          ParameterOptimizeInfo   Object containing information about the iteration.
        """
        if initial_guess is None:
            initial_guess = self.new_problem(frustration=f_func(0), current_sources=Is_func(0)).approximate(algorithm=1)

        if isinstance(initial_guess, StaticConfiguration):
            initial_guess = initial_guess._th()

        out = compute_maximal_parameter(self.get_circuit(), Is_func, f_func, z=0, n=self._nt(),
                                        cp=self.get_current_phase_relation(), theta_0=initial_guess,
                                        lambda_tol=lambda_tol, lambda_initial_stepsize=estimated_upper_bound,
                                        stepsize_reduction_factor=DEF_MAX_PAR_REDUCE_FACT,
                                        newton_tol=newton_tol, newton_maxiter=newton_maxiter,
                                        newton_stop_as_residual_increases=newton_stop_as_residual_increases,
                                        newton_stop_if_not_target_n=newton_stop_if_not_target_n, require_stability=require_stability,
                                        stable_maxiter=stable_maxiter, Asq_factorized=self._Asq_factorization())
        lower_bound, upper_bound, theta, info = out

        if lower_bound is None:
            config = None
        else:
            out_problem = self.new_problem(current_sources=Is_func(lower_bound), frustration=f_func(lower_bound))
            config = StaticConfiguration(out_problem, theta)
        return lower_bound, upper_bound, config, info

    def compute_frustration_bounds(self, initial_guess: StaticConfiguration = None,
                                   start_frustration=None, lambda_tol=DEF_MAX_PAR_TOL,
                                   newton_tol=DEF_TOL,
                                   newton_maxiter=DEF_NEWTON_MAXITER,
                                   newton_stop_as_residual_increases=True,
                                   newton_stop_if_not_target_n=False, require_stability=True,
                                   stable_maxiter=DEF_STAB_MAXITER):

        """
        Computes smallest and largest uniform frustration for which a stable solution exists at the
        specified target vortex configuration and source current.

        Input:
        ------
        start_frustration=None                  frustration factor somewhere in the middle of range.
                                                or None; is estimated based on vortex configuration.
        initial_guess=None                      manual approximation at f=start_frustration
                                                or None; then London approximation is used.
        lambda_tol=DEF_MAX_PAR_TOL              see .compute_maximal_parameter()

        (for newton iteration at given lambda, see .compute() for documentation)
        newton_tol=DEF_TOL
        max_total_newton_steps=DEF_MAX_PAR_NEWT_STEPS
        newton_stop_as_residual_increases=True
        newton_stop_if_not_target_n=False

        (for determining stability of solution at given lambda, see StaticConfiguration.is_stable())
        require_stability=True
        stable_maxiter=DEF_STAB_MAXITER

        Output:
        -------
        (smallest_f, largest_f)                 resulting f range
        (smallest_f_config, largest_f_config)   StaticConfigurations at bounds of range.
        (smallest_f_info, largest_f_info)       ParameterOptimizeInfo objects containing
                                                information about the iterations.
        """
        # TODO: take into account areas.

        if start_frustration is None:
            start_frustration = np.mean(self._nt())
        frustration_initial_stepsize = 1.0
        Is_func = lambda x: self._Is()
        f_func_smallest = lambda x: start_frustration - x
        f_func_largest = lambda x: start_frustration + x
        start_problem = self.new_problem(frustration=start_frustration)
        start_config, status, info = start_problem.compute(initial_guess=initial_guess)
        if not start_config.is_stable_target_solution():
            return (None, None), (None, None), (None, None)
        out = self.compute_maximal_parameter(Is_func, f_func_smallest, initial_guess=start_config._th(),
                                             lambda_tol=lambda_tol, estimated_upper_bound=frustration_initial_stepsize,
                                             newton_tol=newton_tol,
                                             newton_maxiter=newton_maxiter,
                                             newton_stop_as_residual_increases=newton_stop_as_residual_increases,
                                             newton_stop_if_not_target_n=newton_stop_if_not_target_n,
                                             require_stability=require_stability, stable_maxiter=stable_maxiter)
        smallest_factor, _, smallest_f_config, smallest_f_info = out
        smallest_f = f_func_smallest(smallest_factor) if smallest_factor is not None else None
        out = self.compute_maximal_parameter(Is_func, f_func_largest, initial_guess=start_config._th(),
                                             lambda_tol=lambda_tol, estimated_upper_bound=frustration_initial_stepsize,
                                             newton_tol=newton_tol,
                                             newton_maxiter=newton_maxiter,
                                             newton_stop_as_residual_increases=newton_stop_as_residual_increases,
                                             newton_stop_if_not_target_n=newton_stop_if_not_target_n,
                                             require_stability=require_stability, stable_maxiter=stable_maxiter)

        largest_factor, _, largest_f_config, largest_f_info = out
        largest_f = f_func_largest(largest_factor) if largest_factor is not None else None
        return (smallest_f, largest_f), (smallest_f_config, largest_f_config), (smallest_f_info, largest_f_info)

    def compute_maximal_current(self, initial_guess: StaticConfiguration = None,
                                lambda_tol=DEF_MAX_PAR_TOL, newton_tol=DEF_TOL,
                                newton_maxiter=DEF_NEWTON_MAXITER,
                                newton_stop_as_residual_increases=True,
                                newton_stop_if_not_target_n=False,
                                require_stability=True, stable_maxiter=DEF_STAB_MAXITER):

        """
        Computes largest source current for which a stable solution exists at the specified target
        vortex configuration and frustration, where the source current is assumed to be
        max_current_factor * self.get_current_sources().

        Input:
        ------
        see .compute_maximal_parameter() for documentation.

        Output:
        -------
        max_current_factor
        net_sources_current     Net sourced current at max_current_factor.
        out_config              StaticConfiguration of state with maximal current.
        info                    ParameterOptimizeInfo objects containing information about the iterations.

        """
        M, Nj = self.get_circuit()._Mr().matrix(), self.get_circuit()._Nj()
        if np.all(self._Is() == 0):
            raise ValueError("Problem must contain nonzero current sources.")

        Is_per_node = np.abs(M @ self._Is())
        max_super_I_per_node = np.abs(M) @ self.get_circuit()._Ic().diag(True, Nj)
        current_factor_initial_stepsize = 1.0 / np.max(Is_per_node / max_super_I_per_node)
        Is_func = lambda x: x * self._Is()
        f_func = lambda x: self._f()
        out = self.compute_maximal_parameter(Is_func, f_func, initial_guess=initial_guess, lambda_tol=lambda_tol,
                                             estimated_upper_bound=current_factor_initial_stepsize,
                                             newton_maxiter=newton_maxiter, newton_tol=newton_tol,
                                             newton_stop_as_residual_increases=newton_stop_as_residual_increases,
                                             newton_stop_if_not_target_n=newton_stop_if_not_target_n,
                                             require_stability=require_stability, stable_maxiter=stable_maxiter)
        max_current_factor, upper_bound, out_config, info = out
        net_I = out_config.get_problem().get_net_sourced_current() if out_config is not None else None
        return max_current_factor, net_I, out_config, info


    def compute_stable_region(self, angles=np.linspace(0, 2*np.pi, 61), start_frustration=None,
                              start_initial_guess: StaticConfiguration | np.ndarray = None,
                              lambda_tol=DEF_MAX_PAR_TOL, newton_tol=DEF_TOL,
                              newton_maxiter=DEF_NEWTON_MAXITER,
                              newton_stop_as_residual_increases=True,
                              newton_stop_if_not_target_n=False,
                              require_stability=True, stable_maxiter=DEF_STAB_MAXITER):

        """
        Finds edge of stable region in (f, Is) space for vortex configuration n.

        Input:
        ------
        angles=np.linspace(0, 2*np.pi, 60)   Angles at which an extremum in (f, Is) space is searched for.
        (see documentation of .compute_maximal_parameter() for other input.)

        Output:
        -------
        frustration  (num_angles,) ndarray          net extermum frustration at each angle
        net_current  (num_angles,) ndarray          net extremum sourced current at each angle
        all_configs  list of StaticConfiguration    config at extreme value for each angle
        all_infos    list of ParameterOptimizeInfo  objects containing information about the iterations at each angle.

        """
        import matplotlib.pyplot as plt

        num_angles = len(angles)

        frust_bnd_prb = self.new_problem(current_sources=0)
        out = frust_bnd_prb.compute_frustration_bounds(initial_guess=start_initial_guess,
                                              start_frustration=start_frustration,
                                              lambda_tol=lambda_tol,
                                              newton_tol=newton_tol, newton_maxiter=newton_maxiter,
                                              newton_stop_as_residual_increases=newton_stop_as_residual_increases,
                                              newton_stop_if_not_target_n=newton_stop_if_not_target_n,
                                              require_stability=require_stability, stable_maxiter=stable_maxiter)

        (smallest_f, largest_f), _, _ = out
        if smallest_f is None:
            return None, None, None, None
        dome_center_f = 0.5 * (smallest_f + largest_f)
        dome_center_problem = self.new_problem(frustration=dome_center_f)
        out = dome_center_problem.compute_maximal_current(lambda_tol=lambda_tol,
                                                          newton_tol=newton_tol, newton_maxiter=newton_maxiter,
                                                          newton_stop_as_residual_increases=newton_stop_as_residual_increases,
                                                          newton_stop_if_not_target_n=newton_stop_if_not_target_n,
                                                          require_stability=require_stability, stable_maxiter=stable_maxiter)
        max_current_factor, _, _, _ = out
        if max_current_factor is None:
            return None, None, None, None
        frustration = np.zeros(num_angles, dtype=np.double)
        net_current = np.zeros(num_angles, dtype=np.double)
        all_configs, all_infos = [], []
        dome_center_problem = self.new_problem(frustration=dome_center_f, current_sources=0)
        for angle_nr in range(num_angles):
            angle = angles[angle_nr]
            Is_func = lambda x: x * self._Is() * np.sin(angle) * max_current_factor
            f_func = lambda x: dome_center_f + x * np.cos(angle) * (0.5 * (largest_f - smallest_f))
            out = dome_center_problem.compute_maximal_parameter(Is_func, f_func, lambda_tol=lambda_tol,
                                                 newton_tol=newton_tol, newton_maxiter=newton_maxiter,
                                                 newton_stop_as_residual_increases=newton_stop_as_residual_increases,
                                                 newton_stop_if_not_target_n=newton_stop_if_not_target_n,
                                                 require_stability=require_stability, stable_maxiter=stable_maxiter)
            lower_bound, upper_bound, out_config, info = out
            net_current[angle_nr] = out_config.get_problem().get_net_sourced_current() * np.sign(np.sin(angle)) if lower_bound is not None else np.nan
            frustration[angle_nr] = f_func(lower_bound) if lower_bound is not None else np.nan
            all_configs += [out_config]
            all_infos += [info]

        return frustration, net_current, all_configs, all_infos

    def __str__(self):
        return "static problem: " + \
               "\n\tcurrent sources: " + self.get_current_sources().__str__() + \
               "\n\tfrustration: " + self.get_frustration().__str__() + \
               "\n\tvortex configuration: " + self.get_vortex_configuration().__str__() + \
               "\n\tphase zone: " + self.get_phase_zone().__str__() + \
               "\n\tcurrent-phase relation: " + self.current_phase_relation.__str__()

    def _Is(self):
        return self.current_sources

    def _f(self):
        return self.frustration

    def _nt(self):
        return self.vortex_configuration

    def _cp(self, Ic: Matrix, theta):
        return self.current_phase_relation.eval(Ic.diag(), theta)

    def _dcp(self, Ic: Matrix, theta):
        return self.current_phase_relation.d_eval(Ic.diag(), theta)

    def _icp(self, Ic: Matrix, theta):
        return self.current_phase_relation.i_eval(Ic.diag(), theta)

    def _Is_norm(self):
        if self.current_sources_norm is None:
            self.current_sources_norm = scipy.linalg.norm(np.broadcast_to(self.current_sources, (self.circuit._Nj(),)))
        return self.current_sources_norm

    def _Asq_factorization(self):
        if self.Asq_factorization is None:
            A = self.get_circuit().get_cycle_matrix()
            self.Asq_factorization = scipy.sparse.linalg.factorized(A @ A.T)
        return self.Asq_factorization

    def _AIpLIcA_factorization(self):
        if self.AIpLIcA_factorization is None:
            Nj, A = self.get_circuit()._Nj(), self.get_circuit().get_cycle_matrix()
            L, Ic = self.get_circuit()._L().matrix(Nj), self.get_circuit()._Ic().matrix(Nj)
            self.AIpLIcA_factorization = scipy.sparse.linalg.factorized(A @ (scipy.sparse.eye(Nj) + L @ Ic) @ A.T)
        return self.AIpLIcA_factorization

    def _IpLIc_factorization(self):
        if self.IpLIc_factorization is None:
            Nj = self.get_circuit()._Nj()
            L, Ic = self.get_circuit()._L().matrix(Nj), self.get_circuit()._Ic().matrix(Nj)
            self.IpLIc_factorization = scipy.sparse.linalg.factorized(scipy.sparse.eye(Nj) + L @ Ic)
        return self.IpLIc_factorization

    def _Msq_factorization(self):
        if self.Msq_factorization is None:
            M = self.get_circuit()._Mr().A
            self.Msq_factorization = scipy.sparse.linalg.factorized(M @ M.T)
        return self.Msq_factorization

class StaticConfiguration:
    """
    Approximation or solution to static problem.

    It is defined by a StaticProblem and theta. Here theta must be a
    numpy array of shape (Nj,).

    Methods:
    --------
    get_problem()                    Returns problem
    get_circuit()                      Returns circuit (stored in problem)
    get_phi()                        Returns (Nn,) array containing phases at each node
    get_theta()                      Returns (Nj,) array containing gauge invariant phase difference at each junction
    get_n()                          Returns (Nf,) int array containing vorticity at each face
    get_I()                          Returns (Nj,) array containing current through each junction
    get_J()                          Returns (Nf,) array containing path current around each face
    get_flux()                       Returns (Nf,) array containing magnetic flux at each face
    get_EM()                         Returns (Nj,) array containing magnetic energy at each junction
    get_EJ()                         Returns (Nj,) array containing Josephson energy at each junction
    get_Etot()                       Returns get_EM() + get_EJ()
    satisfies_kirchhoff_rules()      Returns if configuration satisfies kirchhoff's rules
    satisfies_winding_rules()        Returns if configuration satisfies the winding rules
    satisfies_target_vortices()      Returns if vortex configuration equals that of problem
    is_stable(maxiter, accept_ratio) Returns if configuration is dynamically stable.
    is_solution(tol)                 Returns if configuration is a solution meaning it must satisfy both kirchhoff
                                     and winding rules
    is_target_solution(tol)          is_solution() and satisfies_target_vortices()
    is_stable_target_solution()      is_solution() and satisfies_target_vortices() and is_stable()
    get_error_kirchhoff_rules()      Returns normalized residual of kirchhoff's rules (normalized so cannot exceed 1)
    get_error_winding_rules()        Returns normalized residual of the winding rules (normalized so cannot exceed 1)
    get_error()                      Returns get_error_kirchhoff_rules(), get_error_winding_rules()
    plot(...) -> plot_handles        Creates plot of configuration.
    """

    def __init__(self, problem: StaticProblem, theta: np.ndarray):
        self.problem = problem
        self.theta = np.array(theta)
        if not self.theta.shape == (self.problem.get_circuit()._Nj(),):
            raise ValueError("theta must be of shape (Nj,)")

    def get_circuit(self) -> Circuit:
        return self.problem.get_circuit()

    def get_problem(self) -> StaticProblem:
        return self.problem

    def get_phi(self) -> np.ndarray:
        # by default the last node (node with highest index number) is grounded.
        M, Msq_solver = self.get_circuit()._Mr().matrix(), self.get_problem()._Msq_factorization()
        return np.append(Msq_solver(M @ self._th()), [0])

    def get_theta(self) -> np.ndarray:
        return self.theta

    def get_n(self) -> np.ndarray:
        A, tpr = self.get_circuit().get_cycle_matrix(), 1.0 / (2.0 * np.pi)
        return - (A @ np.round(self._th() / (2.0 * np.pi))).astype(int)

    def get_I(self) -> np.ndarray:
        return self.problem._cp(self.get_circuit()._Ic(), self._th())

    def get_J(self) -> np.ndarray:
        A, Asq_solver = self.get_circuit().get_cycle_matrix(), self.get_problem()._Asq_factorization()
        return Asq_solver(A @ self.get_I())

    def get_flux(self) -> np.ndarray:
        A, Nj = self.get_circuit().get_cycle_matrix(), self.get_circuit()._Nj()
        return self.problem.frustration + A @ self.get_circuit()._L().matrix(Nj) @ self.get_I() / (2 * np.pi)

    def get_EM(self) -> np.ndarray:
        Nj = self.get_circuit()._Nj()
        return 0.5 * self.get_circuit()._L().matrix(Nj) @ (self.get_I() ** 2)

    def get_EJ(self) -> np.ndarray:
        return self.problem._icp(self.get_circuit()._Ic(), self._th())

    def get_Etot(self) -> np.ndarray:
        return self.get_EJ() + self.get_EM()

    def satisfies_kirchhoff_rules(self, tol=DEF_TOL):
        return self.get_error_kirchhoff_rules() < tol

    def satisfies_winding_rules(self, tol=DEF_TOL):
        return self.get_error_winding_rules() < tol

    def satisfies_target_vortices(self):
        return np.all(self.get_n() == self.problem.get_vortex_configuration())

    def is_stable(self, maxiter=DEF_STAB_MAXITER) -> bool:
        """
        Determines if a configuration is dynamically stable by determining if the Jacobian matrix
        of the time-evolution at the stationairy point is negative definite.

        This is done using the scipy implementation of the lobpcg method.
         - This iterative method converges to the largest eigenvalue (from below).
         - It is prematurely cut-off if the eigenvalue l at iteration i:
           * l(i) > 0 (hopeless)
           * l(i) + accept_ratio * tolerance(i) < 0 ("judgement call" lambda will never exceed 0)

        The last can be rephrased as that it stops when the ratio -l(i)/tolerance(i) exceeds accept_tol
        and is "accepted".

        Inputs:
        -------
        maxiter=DEF_STAB_MAXITER        maximum number of iterations to determine if solutions are stable
        accept_ratio=DEF_STAB_RATIO     residual-to-max-eigenvalue ratio above which is decided max-eigv
                                        must be negative.
        Outputs:
        --------
        is_stable               True if configuration is dynamically stable
        stable_iter_info        StableIterInfo object containing numerical info about the lobpcg iteration
        """
        cp = self.get_problem().get_current_phase_relation()
        out_is_stable = is_stable(self.get_circuit(), self._th(), cp, maxiter=maxiter)
        return out_is_stable

    def is_solution(self, tol=DEF_TOL):
        return self.satisfies_kirchhoff_rules(tol) & self.satisfies_winding_rules(tol)

    def is_target_solution(self, tol=DEF_TOL):
        return self.is_solution(tol=tol) & self.satisfies_target_vortices()

    def is_stable_target_solution(self, tol=DEF_TOL, stable_maxiter=DEF_STAB_MAXITER):
        return self.is_target_solution(tol=tol) & self.is_stable(maxiter=stable_maxiter)

    def get_error_kirchhoff_rules(self) -> np.ndarray:
        return get_kirchhoff_error(self.get_circuit(), self.get_I(), self.get_problem()._Is(),
                                   precomputed_Is_norm=self.problem._Is_norm())

    def get_error_winding_rules(self) -> np.ndarray:
        circuit, problem = self.get_circuit(), self.get_problem()
        f, Asq_factorized = problem._f(),  problem._Asq_factorization()
        L = circuit._L().matrix(self.get_circuit()._Nj())
        return get_winding_error(circuit, self._th() + L @ self.get_I(), get_g(circuit, f, 0, Asq_solver=Asq_factorized))

    def get_error(self):
        return self.get_error_kirchhoff_rules(), self.get_error_winding_rules()

    def plot(self, show_vortices=True, vortex_diameter=0.25, vortex_color=(0, 0, 0),
             anti_vortex_color=(0.8, 0.1, 0.2), vortex_alpha=1, show_grid=True, grid_width=1,
             grid_color=(0.3, 0.5, 0.9), grid_alpha=0.5, show_colorbar=True, show_arrows=True,
             arrow_quantity="I", arrow_width=0.005, arrow_scale=1, arrow_headwidth=3, arrow_headlength=5,
             arrow_headaxislength=4.5, arrow_minshaft=1, arrow_minlength=1, arrow_color=(0.2, 0.4, 0.7),
             arrow_alpha=1, show_nodes=True, node_diameter=0.2,
             node_face_color=(1, 1, 1), node_edge_color=(0, 0, 0), node_alpha=1, show_node_quantity=False,
             node_quantity="phase", node_quantity_cmap=None, node_quantity_clim=(0, 1), node_quantity_alpha=1,
             node_quantity_logarithmic_colors=False, show_face_quantity=False, face_quantity="n",
             face_quantity_cmap=None, face_quantity_clim=(0, 1), face_quantity_alpha=1,
             face_quantity_logarithmic_colors=False,
             figsize=None, title="", **kwargs):
        """
        See CircuitPlot for documentation.
        """
        from pyJJAsim.circuit_visualize import CircuitPlot

        return CircuitPlot(self, show_vortices=show_vortices, vortex_diameter=vortex_diameter,
                         vortex_color=vortex_color, anti_vortex_color=anti_vortex_color,
                         vortex_alpha=vortex_alpha, show_grid=show_grid, grid_width=grid_width,
                         grid_color=grid_color, grid_alpha=grid_alpha, show_colorbar=show_colorbar,
                         show_arrows=show_arrows,
                         arrow_quantity=arrow_quantity, arrow_width=arrow_width, arrow_scale=arrow_scale,
                         arrow_headwidth=arrow_headwidth, arrow_headlength=arrow_headlength,
                         arrow_headaxislength=arrow_headaxislength, arrow_minshaft=arrow_minshaft,
                         arrow_minlength=arrow_minlength, arrow_color=arrow_color,
                         arrow_alpha=arrow_alpha,
                         show_nodes=show_nodes, node_diameter=node_diameter,
                         node_face_color=node_face_color, node_edge_color=node_edge_color,
                         node_alpha=node_alpha, show_node_quantity=show_node_quantity,
                         node_quantity=node_quantity, node_quantity_cmap=node_quantity_cmap,
                         node_quantity_clim=node_quantity_clim, node_quantity_alpha=node_quantity_alpha,
                         node_quantity_logarithmic_colors=node_quantity_logarithmic_colors,
                         show_face_quantity=show_face_quantity, face_quantity=face_quantity,
                         face_quantity_cmap=face_quantity_cmap, face_quantity_clim=face_quantity_clim,
                         face_quantity_alpha=face_quantity_alpha,
                         face_quantity_logarithmic_colors=face_quantity_logarithmic_colors,
                         figsize=figsize, title=title, **kwargs).make()

    def report(self):
        print("Kirchhoff rules error:    ", self.get_error_kirchhoff_rules())
        print("Path rules error:         ", self.get_error_winding_rules())
        print("is stable:                ", self.is_stable())
        print("is target vortex solution:", self.satisfies_target_vortices())

    def _th(self):
        return self.theta


"""
UTILITY ALGORITHMS
"""

def get_kirchhoff_error(circuit: Circuit, I, Is, precomputed_Is_norm=None):
    # Residual of kirchhoffs current law: M @ (I - Is) = 0. Normalized; so between 0 and 1.
    if precomputed_Is_norm is None:
        precomputed_Is_norm = scipy.linalg.norm(Is)
    b = circuit.get_cut_matrix() @ (I - Is)
    M_norm = circuit.get_M_norm()
    normalizer = M_norm * (precomputed_Is_norm + scipy.linalg.norm(I))
    return np.finfo(float).eps if np.abs(normalizer) < 1E-20 else scipy.linalg.norm(b) / normalizer

def get_winding_error(circuit: Circuit, th_p, g):
    # Residual of winding rule: A @ (thp - g) = 0. Normalized; so between 0 and 1. (where thp = th + L @ I)
    A = circuit.get_cycle_matrix()
    A_norm = circuit.get_A_norm()
    normalizer = A_norm * (scipy.linalg.norm(th_p) + scipy.linalg.norm(g))
    return np.finfo(float).eps if np.abs(normalizer) < 1E-20 else scipy.linalg.norm(A @ (th_p - g)) / normalizer

def principle_value(theta):
    return theta - 2 * np.pi * np.round(theta / (2 * np.pi))

def get_g(circuit: Circuit, f=0, z=0, Asq_solver=None):
    # g vector obeying A @ g = 2 * pi * (z - f)
    A, Nf = circuit.get_cycle_matrix(), circuit._Nf()
    if Asq_solver is None:
        Asq_solver = scipy.sparse.linalg.factorized(A @ A.T)
    return 2 * np.pi * A.T @ Asq_solver(np.broadcast_to(z - f, (Nf,)))

def change_phase_zone(circuit: Circuit, theta, z_old, z_new):
    # adds multiples of 2*pi to theta such that it obeys A @ (th_new + L @ I) = 2 * pi * (z_new - f)
    # (assuming it already satisfied A @ (th_old + L @ I) = 2 * pi * (z_old- f))
    return theta + circuit._A_solve(np.broadcast_to(z_new - z_old, (circuit._Nf(),)).copy()) * 2.0 * np.pi

def node_to_junction_current(circuit: Circuit, node_current):
    Mr = circuit._Mr().A
    return -Mr.T @ scipy.sparse.linalg.spsolve(Mr @ Mr.T, node_current[:-1])

"""
PARAMETER MAXIMIZATION ALGORITHMS
"""

def compute_maximal_parameter(circuit: Circuit, Is_func, f_func, z, n, cp=DefaultCPR(),
                              theta_0=None, lambda_tol=DEF_MAX_PAR_TOL, lambda_initial_stepsize=1.0,
                              stepsize_reduction_factor=DEF_MAX_PAR_REDUCE_FACT, newton_tol=DEF_TOL,
                              newton_maxiter=DEF_NEWTON_MAXITER, newton_stop_as_residual_increases=True,
                              newton_stop_if_not_target_n=False, require_stability=True,
                              stable_maxiter=DEF_STAB_MAXITER, Asq_factorized=None):

    """
    Core algorithm for parameter optimization.

    Stand-alone method. The wrappers StaticProblem and StaticConfiguration are more convenient.

    Finds the largest value of lambda for which a problem which current_sources=Is_func(lambda)
    and frustration=f_func(lambda) has a stable stationary state.

     - Must be able to find a stable configuration at lambda=0.
     - One can manually specify an initial_guess for lambda=0.
     - returns a lower- and upperbound for lambda. Stops when the difference < lambda_tol * lower_bound
     - furthermore returns config containing the solutions at the lower_bound. Also its
       accompanied problem has f and Is of lower_bound.
     - ignores self.current_sources and self.frustration.
     - Also returns ParameterOptimizeInfo object containing information about the iteration.
     - Algorithm stops if lambda_tol is reached or when newton_iteration failed to converge or diverge.

    Input:
    ------
    circuit                                   Circuit object
    Is_func                                 Is_func(lambda) -> Is
    f_func                                  f_func(lambda) -> f
    z                                       (Nf,) array containing phase zone
    n                                       (Nf,) array containing vortex configuration
    cp=DefaultCPR()        current-phase relation
    theta_0=None                            manual approximation at lambda=0
                                            or None; then London approximation is used.
    lambda_tol=DEF_MAX_PAR_TOL              stop iterating if upperbound - lowerbound < lambda_tol * lower_bound
    lambda_initial_stepsize=1.0             Estimate for the upperbound for lambda used as initial stepsize

    (for newton iteration at given lambda, see .compute() for documentation)
    newton_tol=DEF_TOL
    max_total_newton_steps=DEF_MAX_PAR_NEWT_STEPS
    newton_stop_as_residual_increases=True
    newton_stop_if_not_target_n=False

    (for determining stability of solution at given lambda, see StaticConfiguration.is_stable())
    require_stability=True,
    stable_maxiter=DEF_STAB_MAXITER
    Asq_factorized                      Solver for A @ A.T @ x == b. If None, set equal to
                                        scipy.sparse.linalg.factorized(A @ A.T)
    Output:
    -------
    lambda_lowerbound       float                   Lowerbound of lambda
    lambda_upperbound       float                   Upperbound of lambda
    theta0                  ndarray                 solution at lambda=lambda_lowerbound
    iteration_info          ParameterOptimizeInfo   Object containing information about the iteration.
    """

    # prepare info handle
    info = ParameterOptimizeInfo(Is_func, f_func, lambda_tol, circuit.get_cut_matrix())

    # prepare matrices
    if Asq_factorized is None:
        A = circuit.get_cycle_matrix()
        Asq_factorized = scipy.sparse.linalg.factorized(A @ A.T)

    # determine solution at lambda=0
    Is, f = Is_func(0), f_func(0)
    if theta_0 is None:
        theta_0 = london_approximation(circuit, f, n)
        theta_0 = change_phase_zone(circuit, theta_0, n, z)
    out = static_compute(circuit, theta_0, Is, f, n, z, cp, tol=newton_tol,
                         maxiter=newton_maxiter, Asq_solver=Asq_factorized,
                         stop_as_residual_increases=newton_stop_as_residual_increases,
                         stop_if_not_target_n=newton_stop_if_not_target_n)
    theta, status, newton_iter_info = out[0], out[1], out[2]
    is_solution = newton_iter_info.found_target_solution()

    if is_solution and require_stability:
        is_solution &= is_stable(circuit, theta, cp, maxiter=stable_maxiter)

    info._preset(is_solution)

    # return if no solution at lambda=0
    if not is_solution:
        return None, None, None, info

    # prepare iteration to find maximum lambda
    found_upper_bound = False
    lambda_stepsize = lambda_initial_stepsize
    lambda_val = lambda_stepsize
    theta0 = theta

    # start iteration to find maximum lambda
    iter_nr = 0
    while True:

        # determine solution at current lambda
        Is, f = Is_func(lambda_val), f_func(lambda_val)
        out = static_compute(circuit, theta_0, Is, f, n, z, cp, tol=newton_tol,
                             maxiter=newton_maxiter, Asq_solver=Asq_factorized,
                             stop_as_residual_increases=newton_stop_as_residual_increases,
                             stop_if_not_target_n=newton_stop_if_not_target_n)
        theta, status, newton_iter_info = out[0], out[1],  out[2]
        is_solution = newton_iter_info.found_target_solution()
        if status == 2:
            break

        if require_stability and is_solution:
            is_solution &= is_stable(circuit, theta, cp, maxiter=stable_maxiter)

        # update information on current iteration in info handle
        info._set(lambda_val, lambda_stepsize, is_solution, newton_iter_info)

        # determine new lambda value to try (and corresponding initial condition)
        if is_solution:
            lambda_val += lambda_stepsize
            theta0 = theta
        else:
            lambda_val -= lambda_stepsize * (1 - stepsize_reduction_factor)
            lambda_stepsize*=stepsize_reduction_factor
            found_upper_bound = True
        if (lambda_stepsize / lambda_val) < lambda_tol:
            break
        iter_nr += 1

    # determine lower- and upperbound on lambda
    info._finish(status)
    lower_bound = lambda_val - lambda_stepsize
    upper_bound = lambda_val if found_upper_bound else np.inf

    return lower_bound, upper_bound, theta0, info


"""
APPROXIMATE STATE FINDING ALGORITHMS
"""

def london_approximation(circuit: Circuit, f, n, AIpLIcA_solver=None):
    A, Nf = circuit.get_cycle_matrix(),  circuit._Nf()
    if AIpLIcA_solver is None:
        Nj = circuit._Nj()
        L, Ic = circuit._L().matrix(Nj), circuit._Ic().matrix(Nj)
        AIpLIcA_solver = scipy.sparse.linalg.factorized(A @ (scipy.sparse.eye(Nj) + L @ Ic) @ A.T)
    return 2 * np.pi * A.T @ AIpLIcA_solver(np.broadcast_to(n - f, (Nf,)))

def arctan_approximation(circuit: Circuit, f, n, Asq_solver=None, IpLIc_solver=None):
    centr_x, centr_y = circuit.get_face_centroids()
    return arctan_approximation_placed_vortices(circuit, f, n[n != 0], centr_x[n != 0], centr_y[n != 0],
                                                Asq_solver=Asq_solver, IpLIc_solver=IpLIc_solver)

def arctan_approximation_placed_vortices(circuit: Circuit, f, n, x_n, y_n, Asq_solver=None, IpLIc_solver=None):
    n = np.atleast_1d(n)
    x_n = np.atleast_1d(x_n)
    y_n = np.atleast_1d(y_n)
    MT = circuit.get_cut_matrix().T
    if IpLIc_solver is None:
        Nj = circuit._Nj()
        L, Ic = circuit._L().matrix(Nj), circuit._Ic().matrix(Nj)
        IpLIc_solver = scipy.sparse.linalg.factorized(scipy.sparse.eye(Nj) + L @ Ic)
    x, y = circuit.get_node_coordinates()
    MTphi = MT @ np.sum(np.arctan2(y - y_n[:, None], x - x_n[:, None]) * n[:, None], axis=0)
    out = principle_value(MTphi) + get_g(circuit, f=f, z=0, Asq_solver=Asq_solver)
    return IpLIc_solver(out) + 2 * np.pi * np.round(MTphi / (2 * np.pi))


"""
STATIONAIRY STATE FINDING ALGORITHMS
"""


def static_compute(circuit: Circuit, theta0=0, Is=0, f=0, n=0, z=0,
                   cp=DefaultCPR(), tol=DEF_TOL,
                   maxiter=DEF_NEWTON_MAXITER, Asq_solver=None,
                   stop_as_residual_increases=True, stop_if_not_target_n=False):
    """
    Core algorithm computing stationary state of a Josephson Junction Circuit using Newtons method.

    Stand-alone method. The wrappers StaticProblem and StaticConfiguration are more convenient.

    Input:
    ------
    circuit                               Circuit                   josephson junction circuit
    theta0                              (Nj,) ndarray           initial guess
    Is=0                                (Nj,) ndarray           Current sources at each junction
    f=0                                 (Nf,) ndarray           Frustration in each face
    n=0                                 (Nf,) ndarray           number of vortices in each face
    z=0                                 (Nf,) ndarray           phase zone of each face
    cp=DefaultCPR()    CurrentPhaseRelation    current phase relation
    tol=DEF_TOL                         scalar                  tolerance. is solution if |residual| < tol
    max_iter=100                        scalar                  maximum number of newton iterations.
    Asq_solver=None                     func: vector->vector    Solver for A @ A.T @ x == b. If None, set equal to
                                                                scipy.sparse.linalg.factorized(A @ A.T)
    stop_as_residual_increases=True     bool                    iteration stops if error(iter) > error(iter - 3)
    stop_if_not_target_n=False          bool                    iteration stops if n != target_n

    Output:
    -------
    theta                              (Nj,) ndarray            gauge invariant phase difference of solution
    convergence_status                 int                      0 -> converged
                                                                1 -> diverged
                                                                2 -> max_iter reached without converging or diverging.
    info                               NewtonIterInfo           Information about iteration (timing, steps, residuals, etc)


    Extra information:
    ------------------
    Stops iterating if:                         convergence_status
     - residual is smaller than tol             0 (converged) if get_n(theta) == n
                                                1 (diverged)  if get_n(theta) != n
     - iteration number iter exceeds maxiter    2 (indeterminate)
     - residual exceeds 0.5                     1 (diverged)
     - if get_n(theta) != n and                 1 (diverged)
       stop_if_not_target_n==True
     - resid(iter) > resid(iter-3) and          1 (diverged)
       stop_as_residual_increases==True

    """

    # prepare newton iter info
    info = NewtonIterInfo(tol, maxiter)

    # get circuit quantities and matrices
    Nj, Nf, M, A = circuit._Nj(), circuit._Nf(), circuit.get_cut_matrix()[:-1, :], circuit.get_cycle_matrix()
    L = circuit._L().matrix(Nj)
    Ic = np.broadcast_to(circuit.get_critical_current_factors(), (Nj,))

    Is = np.ones((Nj,), dtype=np.double) * Is if np.array(Is).size == 1 else Is
    f = np.ones((Nf,), dtype=np.double) * f if np.array(f).size == 1 else f
    n = np.ones((Nf,), dtype=int) * n if np.array(n).size == 1 else n

    # iteration-0 computations
    if Asq_solver is None:
        Asq_solver = scipy.sparse.linalg.factorized(A @ A.T)
    g = get_g(circuit, f=f, z=z, Asq_solver=Asq_solver)

    theta = theta0.copy()
    I = cp.eval(Ic, theta)
    Is_norm = scipy.linalg.norm(Is)

    # iteration-0 compute errors
    error1 = get_kirchhoff_error(circuit, I, Is,  precomputed_Is_norm=Is_norm)
    error2 = get_winding_error(circuit, theta + L @ I, g)

    error = max(error1, error2)
    is_target_n = np.all(A @ (np.round(theta / (2 * np.pi))).astype(int) == z - n)
    info._set(error, is_target_n)

    # prepare newton iteration
    prev_error = np.inf
    iteration = 0


    while not (error < tol or (error > 0.5 and iteration > 5) or (stop_if_not_target_n and is_target_n) or
               (stop_as_residual_increases and error > prev_error) or iteration >= maxiter):
        # iteration computations

        q = cp.d_eval(Ic, theta)
        J = scipy.sparse.vstack([M @ scipy.sparse.diags(q, 0), A @ (scipy.sparse.eye(Nj) + L @ scipy.sparse.diags(q, 0))]).tocsc()
        J_solver = scipy.sparse.linalg.factorized(J)
        F = np.concatenate([M @ (I - Is), A @  (theta - g + L @ I)])
        theta -= J_solver(F)
        I = cp.eval(Ic, theta)

        # iteration error computations
        error1 = get_kirchhoff_error(circuit, I, Is, precomputed_Is_norm=Is_norm)
        error2 = get_winding_error(circuit, theta + L @ I, g)
        error = max(error1, error2)
        is_target_n = np.all(A @ (np.round(theta / (2 * np.pi))).astype(int) == z - n)
        info._set(error, is_target_n)
        if iteration >= 3:
            prev_error = info.error[iteration - 3]

        iteration += 1

    return theta, info.get_status(), info

"""
STABILITY ALGORITHMS
"""

def is_stable(circuit: Circuit, theta, cp, maxiter=DEF_STAB_MAXITER):
    """
    Determines if a configuration on a circuit with inductance is stable in the sense that
    the Jacobian is negative definite. Does not explicitly check if configuration is a stationairy point.
    """
    Nj, Nnr = circuit._Nj(), circuit._Nnr()
    A, M, L = circuit.get_cycle_matrix(),  circuit._Mr().A, circuit._L().matrix(Nj)
    Ic = circuit._Ic().diag(force_as_vector=True, vector_length=Nj)
    q = cp.d_eval(Ic, theta)
    AL = (A @ L @ A.T).tocoo()
    ALL = scipy.sparse.coo_matrix((AL.data, (AL.row + Nnr, AL.col + Nnr)), shape=(Nj, Nj)).tocsc()
    m = scipy.sparse.vstack([M, A @ L]).tocsc()
    J = - (m @ scipy.sparse.diags(q, 0) @ m.T + ALL)
    select = np.diff(J.indptr)!=0
    J = J[select, :][:, select]
    try:
        w, v = scipy.sparse.linalg.eigsh(J, 1, maxiter=maxiter, which="LA")
        out_is_stable = w[0] < 20 * np.finfo(float).eps
    except ArpackNoConvergence:
        print("warning: eigsh ran out of steps")
        out_is_stable = False
    return out_is_stable





from josephson_circuit import Circuit
from static_problem import StaticProblem
from time_evolution import TimeEvolutionProblem, DEF_TEMPERATURE_PROFILE, AnnealingProblem2
import numpy as np
import matplotlib
import matplotlib.pyplot as plt

matplotlib.use("TkAgg")

def find_extreme_vortex_configuration(array: Circuit, Is_base, f, start_lambda=0.5, W=10,
                                      dt=2.0, interval_count=1000, interval_steps=20,
                                      vortex_mobility=0.001, start_T=1, T_factor=1.03,
                                      lambda_factor=0.8, rerun_at_max_lambda=False, _maxtries=3):
    """
    Attempts to find a vortex configuration which allows the largest value of lambda
    """
    Is_func = lambda x: Is_base * x
    f_func = lambda x: f

    print(f"searching at lambda={start_lambda}")
    out_lambda = np.zeros(0, dtype=np.double)
    out_n = np.zeros((array.face_count(), 0), dtype=int)
    cur_f = f_func(start_lambda)
    cur_Is = Is_func(start_lambda)

    # main step
    prob = AnnealingProblem2(array, time_step=dt, interval_steps=interval_steps, interval_count=interval_count,
                             vortex_mobility=vortex_mobility, frustration=cur_f,
                             current_sources=cur_Is, problem_count=W, start_T=start_T, T_factor=T_factor)
    vortex_configurations, energies, status, configurations, T = prob.compute()
    vortex_configurations = vortex_configurations[:, status==0]
    Wp = np.sum(status==0)
    if np.sum(status==0) == 0:
        if _maxtries > 1:
            l, n = find_extreme_vortex_configuration(array, Is_base, f, start_lambda=start_lambda * lambda_factor, W=W,
                                          dt=dt, interval_count=interval_count, interval_steps=interval_steps,
                                          vortex_mobility=vortex_mobility, start_T=start_T, T_factor=T_factor,
                                          lambda_factor=lambda_factor, _maxtries=_maxtries-1, rerun_at_max_lambda=False)
            out_lambda = np.append(out_lambda, l)
            out_n = np.concatenate((out_n, n), axis=1)
    else:
        lambdas = np.zeros(Wp)
        confs = np.zeros(Wp, dtype=object)
        for i in range(Wp):
            n = vortex_configurations[:, i]
            m_prob = StaticProblem(array, vortex_configuration=n, frustration=f_func(0))
            lambdas[i], _, confs[i], _ = m_prob.compute_maximal_parameter(Is_func=Is_func, f_func=f_func)
        mask = ~np.isnan(lambdas)
        out_lambda = np.append(out_lambda, lambdas[mask])
        out_n = np.concatenate((out_n, vortex_configurations[:, mask]), axis=1)
        if sum(mask) == 0:
            if _maxtries > 1:
                l, n = find_extreme_vortex_configuration(array, Is_base, f, start_lambda=start_lambda * lambda_factor, W=W,
                                                         dt=dt, interval_count=interval_count, interval_steps=interval_steps,
                                                         vortex_mobility=vortex_mobility, start_T=start_T, T_factor=T_factor,
                                                         lambda_factor=lambda_factor, _maxtries=_maxtries-1, rerun_at_max_lambda=False)
                out_lambda = np.append(out_lambda, l)
                out_n = np.concatenate((out_n, n), axis=1)
    sorter = np.argsort(out_lambda)[::-1]
    out_lambda, out_n = out_lambda[sorter], out_n[:, sorter]
    if rerun_at_max_lambda:
        l, n = find_extreme_vortex_configuration(array, Is_base, f, start_lambda=out_lambda[0], W=W,
                                                 dt=dt, interval_count=interval_count, interval_steps=interval_steps,
                                                 vortex_mobility=vortex_mobility, start_T=start_T, T_factor=T_factor,
                                                 lambda_factor=lambda_factor, _maxtries=_maxtries, rerun_at_max_lambda=False)
        out_lambda = np.append(out_lambda, l)
        out_n = np.concatenate((out_n, n), axis=1)
        sorter = np.argsort(out_lambda)[::-1]
        out_lambda, out_n = out_lambda[sorter], out_n[:, sorter]
    return out_lambda, out_n



import numpy as np

from josephson_circuit import *
from static_problem import *

import matplotlib
import matplotlib.pyplot as plt

matplotlib.use("TkAgg")

"""
EXAMPLE 3: Vortex with inductance: Flux quantization

"""

# Square array with screening
N = 12
array = SquareArray(N, N)
Llist = [0.01, 0.1, 1, 10]
c = (N-1)/2
f = 0  # frustration
n = np.zeros(array.face_count(), dtype=int)  # target vortex configuration
centre_face_idx = array.locate_faces(c, c)  # locating face idx at coordinate x=(N-1)/2, y=(N-1)/2
n[centre_face_idx] = 1

for i, L in enumerate(Llist):
    array.set_inductance_factors(L)
    out, _, _ = StaticProblem(array, frustration=f, vortex_configuration=n).compute()
    flux = out.get_flux()
    flux_center = np.round(flux[array.locate_faces(c, c)], 4)
    flux_sum = np.round(np.sum(flux), 4)
    out.plot(show_face_quantity=True, face_quantity="flux", face_quantity_logarithmic_colors=True,
             face_quantity_clim=[1E-10, 1], title=f"beta_L={L}, flux centre {flux_center}, sum flux {flux_sum}", arrow_color=[1, 1, 1])


N = 30
array = SquareArray(N, N)
Llist = [0.01, 0.1, 1, 10]
c = (N-1)/2
f = 0  # frustration
n = np.zeros(array.face_count(), dtype=int)  # target vortex configuration
centre_face_idx = array.locate_faces(c, c)  # locating face idx at coordinate x=(N-1)/2, y=(N-1)/2
n[centre_face_idx] = 1


Llist = 10 ** np.linspace(-2, 2, 21)
Phi = np.zeros(len(Llist))
for i, L in enumerate(Llist):
    array.set_inductance_factors(L)
    out, _, _ = StaticProblem(array, vortex_configuration=n).compute()
    Phi[i] = np.sum(out.get_flux())
plt.subplots()
plt.semilogx(Llist, Phi, marker="o", label="total flux")
plt.semilogx(Llist, Llist / (1 + Llist), label="beta_L / (1 + beta_L)")
plt.xlabel("beta_L")
plt.ylabel("total flux")

plt.legend()
plt.show()



from functools import partial
from multiprocessing import Pool

import numpy as np

from find_extreme_vortex_configuration import find_extreme_vortex_configuration
from josephson_circuit import Circuit, SquareArray
from static_problem import StaticProblem, DefaultCPR


class VortexConfigurationEnsemble:

    def __init__(self, array: Circuit, frustrations, vortex_configurations):
        """
        array                       Array object
        frustrations                (ensemble_size,) array
        vortex_configurations       (face_count, ensemble_size,) array
        """
        self.array = array
        self.frustrations = np.array(frustrations, dtype=np.double).ravel()
        self.vortex_configurations = np.array(vortex_configurations, dtype=int)
        if self.vortex_configurations.shape != (self.get_array().face_count(), self.frustrations.size):
            raise ValueError("vortex_configurations must have shape (face_count, ensemble_size,)")
        self.vortex_configurations, idx = np.unique(self.vortex_configurations, axis=1, return_index=True)
        self.frustrations = self.frustrations[idx]

    def get_array(self):
        return self.array

    def get_frustrations(self):
        return self.frustrations

    def get_vortex_configurations(self):
        return self.vortex_configurations

    def ensemble_size(self):
        return len(self.frustrations)

    def get_frustration_bounds(self, title=None):
        W = self.ensemble_size()
        array =self.get_array()
        print(W)
        f_low, f_high = np.zeros(W), np.zeros(W)
        f, n = self.frustrations, self.vortex_configurations
        with Pool(processes=16) as pool:
            out = pool.starmap(compute_frust_bnd_func, [(array, f[i], n[:, i]) for i in range(W)])
        for i, o in enumerate(out):
            f_low[i], f_high[i] = o
            print(i, f_low[i], f_high[i])
        mask = np.isnan(f_low) | np.isnan(f_high)
        f_low = f_low[~mask]
        f_high = f_high[~mask]
        ns = self.vortex_configurations[:, ~mask]
        if title is not None:
            with open(title, 'wb') as ffile:
                np.save(ffile, f_low)
                np.save(ffile, f_high)
                np.save(ffile, ns)
        return f_low, f_high, ns

    def compute_maximal_current(self, current_sources_base, num_frustration_points, f_low=None, f_high=None, ns=None):
        if f_low is None:
            f_low, f_high, ns = self.get_frustration_bounds()
        f = np.linspace(np.min(f_low), np.max(f_high), num_frustration_points)
        max_current_factors = np.zeros(num_frustration_points)
        array = self.get_array()
        for i, fp in enumerate(f):
            print(i)
            ids = np.flatnonzero((fp > f_low) & (fp < f_high))
            max_I_factor = 0
            print(ids)
            print(max_I_factor)
            for k in range(1 + ((len(ids)-1) // 16)):
                batch = np.arange(16 * k, min(16 * k + 16, len(ids)))
                print(len(ids), batch)
                if len(batch) > 0:
                    with Pool(processes=16) as pool:
                        out = pool.starmap(compute_max_I_func, [(array, fp, ns[:, i], current_sources_base,
                                                                 max_I_factor) for i in ids[batch]])
                    print([o for o in out])
                    max_I_factor = max(max_I_factor, np.max(np.array([0] + [o for o in out])))
                    print(max_I_factor)
            max_current_factors[i] = max_I_factor
        return f, max_current_factors

    # def compute_max_energy_without_current_sources(self, num_frustration_points):
    #     f_low, f_high = self.get_frustration_bounds()
    #     f = np.linspace(np.min(f_low), np.max(f_high), num_frustration_points)
    #     max_energy = np.zeros(num_frustration_points)
    #     for i, fp in enumerate(f):
    #         energies = [0]
    #         for j, (fl, fh) in enumerate(zip(f_low, f_high)):
    #             n = self.vortex_configurations[:, j]
    #             if fp > fl and fp < fh:
    #                 prob = StaticProblem(self.get_array(), vortex_configuration=n,
    #                                      frustration=fp)
    #                 config, status, info = prob.compute()
    #                 if status == 0:
    #                     energies += [np.mean(config.get_Etot())]
    #         max_energy[i] = np.max(energies)
    #     return f, max_energy


def generate_vortex_ensemble_with_high_current_sources(array, Is_base, f_list, start_lambda=0.2,
                                                       W=30, dt=2.0, interval_count=1000, interval_steps=20,
                                                       vortex_mobility=0.001, start_T=1.0, T_factor=1.03,
                                                       lambda_factor=0.8, rerun_at_max_lambda=True, title=None):
    with Pool(processes=16) as pool:
        out = pool.starmap(find_extreme_vortex_configuration, [(array, Is_base, f, start_lambda, W, dt,
                                                                interval_count, interval_steps, vortex_mobility,
                                                                start_T, T_factor, lambda_factor,
                                                                rerun_at_max_lambda) for f in f_list])
    ns = [o[1][:, :max(len(o[0]), 5)] for o in out]
    fs = np.repeat(f_list, [o.shape[1] for o in ns])
    ns = np.concatenate(ns, axis=1)
    if title is not None:
        with open(title, 'wb') as ffile:
            np.save(ffile, np.array([start_lambda, W, dt, interval_count, interval_steps,
                                     vortex_mobility, start_T, T_factor, lambda_factor]))
            np.save(ffile, fs)
            np.save(ffile, ns)
    return fs, ns


def compute_frust_bnd_func(array, f, n):
    out, _, _ =  StaticProblem(array, vortex_configuration=n).compute_frustration_bounds(start_frustration=f)
    return out

def compute_max_I_func(array, f, n, Is_base, upper):
    out, _, _, _ = StaticProblem(array, vortex_configuration=n, current_sources=Is_base,
                                 frustration=f).compute_maximal_current(abort_if_below=upper)
    return out


def lobpcg_test_negative_definite(A, B, preconditioner=None, accept_ratio=DEF_STAB_RATIO, maxiter=DEF_STAB_MAXITER,
                                  tol=DEF_TOL):

    """
    Computes if inv(B) @ A is negative definite and returns this, along with the largest eigenvalue
    and residual at the last lobpcg step.

    Uses lobpcg iteratively with increasing maxiter and stops if its output max_eigv and residual:
     * if max_eigv + accept_ratio * residual < eps, considered negative definite
     * if max_eigv > eps, considered not negative definite
     * if total_iters > maxiter, apply first criterion
     * if residual < tol, apply first criterion

    Accepts a preconditioner for A.
    """
    eps = 2 * np.finfo(float).eps

    step_iters = 2
    total_iters = step_iters

    x0 = np.random.rand(A.shape[0], 1)
    lobpcg_out = scipy.sparse.linalg.lobpcg(A, x0, B=B, M=preconditioner, maxiter=step_iters, tol=tol,
                                            retLambdaHistory=True, retResidualNormsHistory=True)
    if len(lobpcg_out) <= 2:
        max_eigenvalue = lobpcg_out[0]
        residual = np.array([0])
    else:
        max_eigenvalue = np.stack(lobpcg_out[2]).ravel()[1:]
        residual = np.stack(lobpcg_out[3]).ravel()
    if max_eigenvalue.size == 0:
        max_eigenvalue = np.array([-np.inf])
    max_eigenvalue = np.array(max_eigenvalue)

    while (max_eigenvalue[-1] + accept_ratio * residual[-1] > eps) and total_iters < maxiter and residual[-1] > tol:
        if max_eigenvalue[-1] > eps or total_iters > maxiter:
            return False, max_eigenvalue, residual
        lobpcg_out = scipy.sparse.linalg.lobpcg(A, lobpcg_out[1], B=B, M=preconditioner, maxiter=step_iters, tol=tol,
                                                retLambdaHistory=True, retResidualNormsHistory=True)
        total_iters += step_iters
        step_iters += 1
        if len(lobpcg_out) <= 2:
            max_eigenvalue = np.append(max_eigenvalue, [lobpcg_out[0]])
            residual = np.append(residual, np.array([0]))
        else:
            new_max_eigv = np.stack(lobpcg_out[2]).ravel()[1:]
            if new_max_eigv.size == 0:
                break
            max_eigenvalue = np.append(max_eigenvalue, new_max_eigv)
            residual = np.append(residual, lobpcg_out[3])

    is_negative_definite = max_eigenvalue[-1] + accept_ratio * residual[-1] < eps
    if isinstance(is_negative_definite, np.ndarray):
        is_negative_definite = is_negative_definite[0]
    return is_negative_definite, max_eigenvalue, residual


def eigsh_test_negative_definite(A, B, maxiter=DEF_STAB_MAXITER):
    eps = 2 * np.finfo(float).eps

    w, v = scipy.sparse.linalg.eigsh(A, 1, M=B, maxiter=maxiter)
    eigv = w[0]
    return eigv < eps, eigv, 0.0

def get_max_eigenvalue(A, B, maxiter=DEF_STAB_MAXITER, is_symmetric=True):
    eps = 2 * np.finfo(float).eps
    if is_symmetric:
        w, v = scipy.sparse.linalg.eigsh(A, 1, M=B, maxiter=maxiter)
    else:
        w, v = scipy.sparse.linalg.eigh(A, 1, M=B, maxiter=maxiter)
    eigv = w[0]
    return eigv < eps, eigv, 0.0

def is_stable_no_inductance(array, theta, cp, maxiter=DEF_STAB_MAXITER, accept_ratio=DEF_STAB_RATIO, tol=DEF_TOL):
    """
    Determines if a configuration on an array with no inductance is stable in the sense that
    the Jacobian is negative definite, which it returns along with a StabilityInfo handle.
    Does not explicitly check if configuration is a stationairy point.
    """
    info = StabilityInfo(maxiter, accept_ratio)
    Nj, Nnr = array._Nj(), array._Nnr()
    Ic = array._Ic().diag(force_as_vector=True, vector_length=Nj)
    d_theta = cp.d_eval(Ic, theta)
    M = array._Mr().A
    MT = M.T
    M_sw = -M @ scipy.sparse.diags(d_theta, 0) @ MT
    preconditioner = scipy.sparse.linalg.LinearOperator((Nnr, Nnr), scipy.sparse.linalg.factorized(M_sw))
    B = M @ scipy.sparse.diags(array._R().diag(force_as_vector=True, vector_length=Nj) ** -1, 0) @ MT
    is_stable, max_eigenvalue, residual = lobpcg_test_negative_definite(M_sw, B, accept_ratio=accept_ratio,
                                                                        maxiter=maxiter, preconditioner=preconditioner,
                                                                        tol=tol)
    return is_stable, info._set(max_eigenvalue, residual)

def is_stable_inductance(array: Array, theta, cp, ALA_solver=None, maxiter=DEF_STAB_MAXITER, accept_ratio=5, tol=DEF_TOL):
    """
    Determines if a configuration on an array with inductance is stable in the sense that
    the Jacobian is negative definite, which it returns along with a StabilityInfo handle.
    Does not explicitly check if configuration is a stationairy point.
    """
    info = StabilityInfo(maxiter, accept_ratio)
    A = array.get_cycle_matrix()
    AT = A.T
    Nj = array.junction_count()
    if ALA_solver is None:
        L = array._L().matrix(Nj)
        ALA_solver = scipy.sparse.linalg.factorized(A @ L @ AT)
    d_theta = cp.d_eval(array._Ic().diag(force_as_vector=True, vector_length=array._Nj()), theta)[:, None]
    B = scipy.sparse.diags(array._R().diag(force_as_vector=True, vector_length=array._Nj()) ** -1, 0)
    def func(x):
        x = x.reshape(Nj, 1)
        return -d_theta * x - AT @ ALA_solver(A @ x)
    A_lin = scipy.sparse.linalg.LinearOperator((array._Nj(), array._Nj()), matvec=func)
    is_stable, max_eigenvalue, residual = lobpcg_test_negative_definite(A_lin, B, accept_ratio=accept_ratio, maxiter=maxiter, tol=tol)
    return is_stable, info._set(max_eigenvalue, residual)


class StabilityInfo:

    """
    Information about numerical procedure used to determine stability of a configuration.
    Use print(stability_info) to display the information.

    Methods:
    -------
    get_maxiter()                       Returns maximum number of lobpcg iterations before abort.
    get_accept_ratio()                  Returns eigenvalue_to_residual_ratio at which lobpcg iteration
                                        is aborted because residual is sufficiently below 0 and considered stable.
    get_max_eigenvalue()                Returns maximum eigenvalue found by lobpcg iteration.
    get_iter_count()                    Returns number of lobpcg iterations.
    get_residual()                      Returns residual of last step of lobpcg iteration.
    is_stable()                         Returns if configuration was found to be stable.
    get_runtime()                       Returns runtime in seconds.
    get_eigenvalue_to_residual_ratio()  -returns get_max_eigenvalue() / get_residual()
    """

    def __init__(self,  maxiter, accept_ratio):
        self.maxiter = maxiter
        self.accept_ratio = accept_ratio
        self.start_time = time.perf_counter()
        self.iter_count = 0
        self.residual = np.array(0)
        self.max_eigenvalue = np.array(0)
        self.is_stable_v =  False
        self.elapsed_time = 0

    def get_maxiter(self):
        return self.maxiter

    def get_accept_ratio(self):
        return self.accept_ratio

    def get_max_eigenvalue(self):
        return self.max_eigenvalue[-1]

    def get_iter_count(self):
        return self.iter_count

    def get_residual(self):
        return self.residual[-1]

    def is_stable(self):
        return self.is_stable_v

    def get_runtime(self):
        return self.elapsed_time

    def get_eigenvalue_to_residual_ratio(self):
        return -self.get_max_eigenvalue()/self.get_residual()

    def __str__(self):
        out = f"stability check info: (maxiter={self.get_maxiter()}, accept_ratio={self.get_accept_ratio()})\n\t"
        out += f"is stable: {self.is_stable()}\n\t"
        out += f"step count (of lobpcg method): {self.get_iter_count()}\n\t"
        out += f"maximum eigenvalue: {self.get_max_eigenvalue()}\n\t"
        out += f"lobpcg residual: {self.get_residual()}\n\t"
        out += f"max_eigv-to-residual ratio: {self.get_eigenvalue_to_residual_ratio()}\n\t"
        out += f"elapsed time: {self.get_runtime()} sec"
        return out

    def _set(self, max_eigenvalue, residual):
        self.iter_count = np.array(residual).size
        self.residual = np.array(residual)
        self.max_eigenvalue = np.array(max_eigenvalue)
        self.is_stable_v = max_eigenvalue[-1] < (2 * np.finfo(float).eps)
        self.elapsed_time = time.perf_counter() - self.start_time
        return self

        # if array._has_capacitance():
        # else:
        #     Cnext = Rv
        #     Asq_fact = scipy.sparse.linalg.factorized(A @ scipy.sparse.diags(1.0 / Cnext[:, 0], 0) @ AT)
        #     for i in range(self._Nt()):
        #         Is, T, theta_s, f = self._Is(i), self._T(i), self._theta_s(i), self._f(i)
        #
        #         rand = np.random.randn(Nj, W) if i % 3 == 0 else rand[np.random.permutation(Nj), :]
        #         fluctuations = ((2.0 * T * Rv) ** 0.5) * rand
        #
        #         theta = theta_next.copy()
        #         if array._has_inductance():
        #             y = AT @ L_sw_fact(A @ (theta + theta_s + L @ Is) + 2 * np.pi * f)
        #             theta_next = theta - (self._cp(theta) + fluctuations - Is + y) / Cnext
        #             if self.store_time_steps[i]:
        #                 out._update([theta_next if store_th else None,
        #                              (theta_next - theta) / dt if store_V else None,
        #                              Is - y if store_I else None])
        #         else:
        #             x = (self._cp(theta) + fluctuations - Is - Cnext * theta) / Cnext
        #             theta_next = -x + (AT @ Asq_fact(A @ (x - theta_s) + 2 * np.pi * f)) / Cnext
        #             if self.store_time_steps[i]:
        #                 out._update([theta_next if self.store_theta else None,
        #                              (theta_next - theta) / dt if store_V else None,
        #                              (x + theta_next) * Cnext + Is if store_I else None])




from pyJJAsim import *

import matplotlib
import matplotlib.pyplot as plt
matplotlib.use("TkAgg")

"""
Annealing on array with hole
"""

def make_hole_array(N, L):
    full_array = SquareArray(N, N)
    x, y = full_array.get_node_coordinates()
    hole_face_ids = (x > (N-1-L)/2) & (x < (N-1+L)/2) & \
                    (y > (N-1-L)/2) & (y < (N-1+L)/2)
    hole_array = full_array.remove_nodes(hole_face_ids.flatten())
    return hole_array

if __name__ == "__main__":

    # define array
    N = 20
    L = 6
    hole_array = make_hole_array(N, L)

    # define problem
    f = 0.03 * hole_array.get_face_areas()
    time_step = 2.0
    interval_steps = 20
    interval_count = 2000
    vortex_mobility = 0.001
    problem_count = 250
    start_T = 1.0
    T_factor = 1.03
    problem = AnnealingProblem(hole_array, time_step=time_step, interval_steps=interval_steps,
                               interval_count=interval_count, vortex_mobility=vortex_mobility,
                               frustration=f, problem_count=problem_count,
                               start_T=start_T, T_factor=T_factor)
    # do time simulation
    out = problem.compute()
    vortex_configurations, energies, status, configurations, T = out

    plt.subplots()
    plt.hist(energies, bins=100)
    plt.xlabel("mean energy per junction")
    plt.ylabel("histogram count")

    # plot temperature evolution
    plt.subplots()
    plt.plot(time_step * interval_steps * np.arange(interval_count), T)
    plt.xlabel("time")
    plt.ylabel("annealing temperature")

    lowest_state = np.argmin(energies)
    configurations[lowest_state].plot(vortex_diameter=0.8, figsize=[8, 8])
    plt.title(f"lowest found state at E={np.round(energies[lowest_state], 5)}")

    mask = energies > (energies[lowest_state] + 1E-7)
    second_lowest = np.flatnonzero(mask)[np.argmin(energies[mask])]
    configurations[second_lowest].plot(vortex_diameter=0.8, figsize=[8, 8])
    plt.title(f"second lowest state at E={np.round(energies[second_lowest], 5)}")

    plt.show()
